# System libs
import os
import time
# import math
import random
import argparse
from distutils.version import LooseVersion
# Numerical libs
import torch
import torch.nn as nn
# Our libs
from config import cfg
from dataset import TrainDataset
from models import ModelBuilder, SegmentationModule
from utils import AverageMeter, parse_devices, setup_logger
from lib.nn import UserScatteredDataParallel, user_scattered_collate, patch_replication_callback


# train one epoch
def train(segmentation_module, iterator, optimizers, history, epoch, cfg):
    batch_time = AverageMeter()
    data_time = AverageMeter()
    ave_total_loss = AverageMeter()
    ave_acc = AverageMeter()

    segmentation_module.train(not cfg.TRAIN.fix_bn)

    # main loop
    tic = time.time()
    for i in range(cfg.TRAIN.epoch_iters):
        # load a batch of data
        batch_data = next(iterator)
        data_time.update(time.time() - tic)
        segmentation_module.zero_grad()

        # adjust learning rate
        cur_iter = i + (epoch - 1) * cfg.TRAIN.epoch_iters
        adjust_learning_rate(optimizers, cur_iter, cfg)

        # forward pass
        loss, acc = segmentation_module(batch_data)
        loss = loss.mean()
        acc = acc.mean()

        # Backward
        loss.backward()
        for optimizer in optimizers:
            optimizer.step()

        # measure elapsed time
        batch_time.update(time.time() - tic)
        tic = time.time()

        # update average loss and acc
        ave_total_loss.update(loss.data.item())
        ave_acc.update(acc.data.item()*100)

        # calculate accuracy, and display
        if i % cfg.TRAIN.disp_iter == 0:
            print('Epoch: [{}][{}/{}], Time: {:.2f}, Data: {:.2f}, '
                  'lr_encoder: {:.6f}, lr_decoder: {:.6f}, '
                  'Accuracy: {:4.2f}, Loss: {:.6f}'
                  .format(epoch, i, cfg.TRAIN.epoch_iters,
                          batch_time.average(), data_time.average(),
                          cfg.TRAIN.running_lr_encoder, cfg.TRAIN.running_lr_decoder,
                          ave_acc.average(), ave_total_loss.average()))

            fractional_epoch = epoch - 1 + 1. * i / cfg.TRAIN.epoch_iters
            history['train']['epoch'].append(fractional_epoch)
            history['train']['loss'].append(loss.data.item())
            history['train']['acc'].append(acc.data.item())


def checkpoint(nets, history, cfg, epoch):
    print('Saving checkpoints...')
    (net_encoder, net_decoder, crit) = nets

    dict_encoder = net_encoder.state_dict()
    dict_decoder = net_decoder.state_dict()

    torch.save(
        history,
        '{}/history_epoch_{}.pth'.format(cfg.DIR, epoch))
    torch.save(
        dict_encoder,
        '{}/encoder_epoch_{}.pth'.format(cfg.DIR, epoch))
    torch.save(
        dict_decoder,
        '{}/decoder_epoch_{}.pth'.format(cfg.DIR, epoch))


def group_weight(module):
    group_decay = []
    group_no_decay = []
    for m in module.modules():
        if isinstance(m, nn.Linear):
            group_decay.append(m.weight)
            if m.bias is not None:
                group_no_decay.append(m.bias)
        elif isinstance(m, nn.modules.conv._ConvNd):
            group_decay.append(m.weight)
            if m.bias is not None:
                group_no_decay.append(m.bias)
        elif isinstance(m, nn.modules.batchnorm._BatchNorm):
            if m.weight is not None:
                group_no_decay.append(m.weight)
            if m.bias is not None:
                group_no_decay.append(m.bias)

    assert len(list(module.parameters())) == len(group_decay) + len(group_no_decay)
    groups = [dict(params=group_decay), dict(params=group_no_decay, weight_decay=.0)]
    return groups


def create_optimizers(nets, cfg):
    (net_encoder, net_decoder, crit) = nets
    optimizer_encoder = torch.optim.SGD(
        group_weight(net_encoder),
        lr=cfg.TRAIN.lr_encoder,
        momentum=cfg.TRAIN.beta1,
        weight_decay=cfg.TRAIN.weight_decay)
    optimizer_decoder = torch.optim.SGD(
        group_weight(net_decoder),
        lr=cfg.TRAIN.lr_decoder,
        momentum=cfg.TRAIN.beta1,
        weight_decay=cfg.TRAIN.weight_decay)
    return (optimizer_encoder, optimizer_decoder)


def adjust_learning_rate(optimizers, cur_iter, cfg):
    scale_running_lr = ((1. - float(cur_iter) / cfg.TRAIN.max_iters) ** cfg.TRAIN.lr_pow)
    cfg.TRAIN.running_lr_encoder = cfg.TRAIN.lr_encoder * scale_running_lr
    cfg.TRAIN.running_lr_decoder = cfg.TRAIN.lr_decoder * scale_running_lr

    (optimizer_encoder, optimizer_decoder) = optimizers
    for param_group in optimizer_encoder.param_groups:
        param_group['lr'] = cfg.TRAIN.running_lr_encoder
    for param_group in optimizer_decoder.param_groups:
        param_group['lr'] = cfg.TRAIN.running_lr_decoder


def main(cfg, gpus):
    # Network Builders
    net_encoder = ModelBuilder.build_encoder(
        arch=cfg.MODEL.arch_encoder.lower(),
        fc_dim=cfg.MODEL.fc_dim,
        weights=cfg.MODEL.weights_encoder)
    net_decoder = ModelBuilder.build_decoder(
        arch=cfg.MODEL.arch_decoder.lower(),
        fc_dim=cfg.MODEL.fc_dim,
        num_class=cfg.DATASET.num_class,
        weights=cfg.MODEL.weights_decoder)

    crit = nn.NLLLoss(ignore_index=-1)

    if cfg.MODEL.arch_decoder.endswith('deepsup'):
        segmentation_module = SegmentationModule(
            net_encoder, net_decoder, crit, cfg.TRAIN.deep_sup_scale)
    else:
        segmentation_module = SegmentationModule(
            net_encoder, net_decoder, crit)

    # Dataset and Loader
    dataset_train = TrainDataset(
        cfg.DATASET.root_dataset,
        cfg.DATASET.list_train,
        cfg.DATASET,
        batch_per_gpu=cfg.TRAIN.batch_size_per_gpu)

    loader_train = torch.utils.data.DataLoader(
        dataset_train,
        batch_size=len(gpus),  # we have modified data_parallel
        shuffle=False,  # we do not use this param
        collate_fn=user_scattered_collate,
        num_workers=cfg.TRAIN.workers,
        drop_last=True,
        pin_memory=True)
    print('1 Epoch = {} iters'.format(cfg.TRAIN.epoch_iters))

    # create loader iterator
    iterator_train = iter(loader_train)

    # load nets into gpu
    if len(gpus) > 1:
        segmentation_module = UserScatteredDataParallel(
            segmentation_module,
            device_ids=gpus)
        # For sync bn
        patch_replication_callback(segmentation_module)
    segmentation_module.cuda()

    # Set up optimizers
    nets = (net_encoder, net_decoder, crit)
    optimizers = create_optimizers(nets, cfg)

    # Main loop
    history = {'train': {'epoch': [], 'loss': [], 'acc': []}}

    for epoch in range(cfg.TRAIN.start_epoch, cfg.TRAIN.num_epoch):
        train(segmentation_module, iterator_train, optimizers, history, epoch+1, cfg)

        # checkpointing
        checkpoint(nets, history, cfg, epoch+1)

    print('Training Done!')


if __name__ == '__main__':
    assert LooseVersion(torch.__version__) >= LooseVersion('0.4.0'), \
        'PyTorch>=0.4.0 is required'

    parser = argparse.ArgumentParser(
        description="PyTorch Semantic Segmentation Training"
    )
    parser.add_argument(
        "--cfg",
        default="config/ade20k-resnet50dilated-ppm_deepsup.yaml",
        metavar="FILE",
        help="path to config file",
        type=str,
    )
    parser.add_argument(
        "--gpus",
        default="0-3",
        help="gpus to use, e.g. 0-3 or 0,1,2,3"
    )
    parser.add_argument(
        "opts",
        help="Modify config options using the command-line",
        default=None,
        nargs=argparse.REMAINDER,
    )
    args = parser.parse_args()

    cfg.merge_from_file(args.cfg)
    cfg.merge_from_list(args.opts)
    # cfg.freeze()

    logger = setup_logger(distributed_rank=0)   # TODO
    logger.info("Loaded configuration file {}".format(args.cfg))
    logger.info("Running with config:\n{}".format(cfg))

    # Output directory
    if not os.path.isdir(cfg.DIR):
        os.makedirs(cfg.DIR)
    logger.info("Outputing checkpoints to: {}".format(cfg.DIR))
    with open(os.path.join(cfg.DIR, 'config.yaml'), 'w') as f:
        f.write("{}".format(cfg))

    # Start from checkpoint
    if cfg.TRAIN.start_epoch > 0:
        cfg.MODEL.weights_encoder = os.path.join(
            cfg.DIR, 'encoder_epoch_{}.pth'.format(cfg.TRAIN.start_epoch))
        cfg.MODEL.weights_decoder = os.path.join(
            cfg.DIR, 'decoder_epoch_{}.pth'.format(cfg.TRAIN.start_epoch))
        assert os.path.exists(cfg.MODEL.weights_encoder) and \
            os.path.exists(cfg.MODEL.weights_decoder), "checkpoint does not exitst!"

    # Parse gpu ids
    gpus = parse_devices(args.gpus)
    gpus = [x.replace('gpu', '') for x in gpus]
    gpus = [int(x) for x in gpus]
    num_gpus = len(gpus)
    cfg.TRAIN.batch_size = num_gpus * cfg.TRAIN.batch_size_per_gpu

    cfg.TRAIN.max_iters = cfg.TRAIN.epoch_iters * cfg.TRAIN.num_epoch
    cfg.TRAIN.running_lr_encoder = cfg.TRAIN.lr_encoder
    cfg.TRAIN.running_lr_decoder = cfg.TRAIN.lr_decoder

    random.seed(cfg.TRAIN.seed)
    torch.manual_seed(cfg.TRAIN.seed)

    main(cfg, gpus)



# -*- coding: utf-8 -*-
"""
/*------------------------------------------------------*
|                         BRAILS                        |
|                                                       |
| Author: Charles Wang,  UC Berkeley, c_w@berkeley.edu  |
|                                                       |
| Date:    1/19/2021                                   |
*------------------------------------------------------*/
"""


stateAbbrs =  {'Alabama': 'AL',
     'Alaska': 'AK',
     'Arizona': 'AZ',
     'Arkansas': 'AR',
     'California': 'CA',
     'Colorado': 'CO',
     'Connecticut': 'CT',
     'Delaware': 'DE',
     'District of Columbia': 'DC',
     'Florida': 'FL',
     'Georgia': 'GA',
     'Hawaii': 'HI',
     'Idaho': 'ID',
     'Illinois': 'IL',
     'Indiana': 'IN',
     'Iowa': 'IA',
     'Kansas': 'KS',
     'Kentucky': 'KY',
     'Louisiana': 'LA',
     'Maine': 'ME',
     'Maryland': 'MD',
     'Massachusetts': 'MA',
     'Michigan': 'MI',
     'Minnesota': 'MN',
     'Mississippi': 'MS',
     'Missouri': 'MO',
     'Montana': 'MT',
     'Nebraska': 'NE',
     'Nevada': 'NV',
     'New Hampshire': 'NH',
     'New Jersey': 'NJ',
     'New Mexico': 'NM',
     'New York': 'NY',
     'North Carolina': 'NC',
     'North Dakota': 'ND',
     'Ohio': 'OH',
     'Oklahoma': 'OK',
     'Oregon': 'OR',
     'Pennsylvania': 'PA',
     'Rhode Island': 'RI',
     'South Carolina': 'SC',
     'South Dakota': 'SD',
     'Tennessee': 'TN',
     'Texas': 'TX',
     'Utah': 'UT',
     'Vermont': 'VT',
     'Virginia': 'VA',
     'Washington': 'WA',
     'West Virginia': 'WV',
     'Wisconsin': 'WI',
     'Wyoming': 'WY'}

stateNames = {'AL': 'Alabama',
     'AK': 'Alaska',
     'AZ': 'Arizona',
     'AR': 'Arkansas',
     'CA': 'California',
     'CO': 'Colorado',
     'CT': 'Connecticut',
     'DE': 'Delaware',
     'DC': 'District of Columbia',
     'FL': 'Florida',
     'GA': 'Georgia',
     'HI': 'Hawaii',
     'ID': 'Idaho',
     'IL': 'Illinois',
     'IN': 'Indiana',
     'IA': 'Iowa',
     'KS': 'Kansas',
     'KY': 'Kentucky',
     'LA': 'Louisiana',
     'ME': 'Maine',
     'MD': 'Maryland',
     'MA': 'Massachusetts',
     'MI': 'Michigan',
     'MN': 'Minnesota',
     'MS': 'Mississippi',
     'MO': 'Missouri',
     'MT': 'Montana',
     'NE': 'Nebraska',
     'NV': 'Nevada',
     'NH': 'New Hampshire',
     'NJ': 'New Jersey',
     'NM': 'New Mexico',
     'NY': 'New York',
     'NC': 'North Carolina',
     'ND': 'North Dakota',
     'OH': 'Ohio',
     'OK': 'Oklahoma',
     'OR': 'Oregon',
     'PA': 'Pennsylvania',
     'RI': 'Rhode Island',
     'SC': 'South Carolina',
     'SD': 'South Dakota',
     'TN': 'Tennessee',
     'TX': 'Texas',
     'UT': 'Utah',
     'VT': 'Vermont',
     'VA': 'Virginia',
     'WA': 'Washington',
     'WV': 'West Virginia',
     'WI': 'Wisconsin',
     'WY': 'Wyoming'}

footprintFiles = {'Alabama': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Alabama.zip',
 'Alaska': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Alaska.zip',
 'Arizona': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Arizona.zip',
 'Arkansas': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Arkansas.zip',
 'California': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/California.zip',
 'Colorado': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Colorado.zip',
 'Connecticut': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Connecticut.zip',
 'Delaware': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Delaware.zip',
 'District Of Columbia': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/DistrictofColumbia.zip',
 'Florida': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Florida.zip',
 'Georgia': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Georgia.zip',
 'Hawaii': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Hawaii.zip',
 'Idaho': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Idaho.zip',
 'Illinois': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Illinois.zip',
 'Indiana': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Indiana.zip',
 'Iowa': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Iowa.zip',
 'Kansas': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Kansas.zip',
 'Kentucky': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Kentucky.zip',
 'Louisiana': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Louisiana.zip',
 'Maine': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Maine.zip',
 'Maryland': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Maryland.zip',
 'Massachusetts': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Massachusetts.zip',
 'Michigan': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Michigan.zip',
 'Minnesota': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Minnesota.zip',
 'Mississippi': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Mississippi.zip',
 'Missouri': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Missouri.zip',
 'Montana': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Montana.zip',
 'Nebraska': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Nebraska.zip',
 'Nevada': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Nevada.zip',
 'New Hampshire': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/NewHampshire.zip',
 'New Jersey': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/NewJersey.zip',
 'New Mexico': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/NewMexico.zip',
 'New York': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/NewYork.zip',
 'North Carolina': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/NorthCarolina.zip',
 'North Dakota': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/NorthDakota.zip',
 'Ohio': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Ohio.zip',
 'Oklahoma': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Oklahoma.zip',
 'Oregon': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Oregon.zip',
 'Pennsylvania': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Pennsylvania.zip',
 'Rhode Island': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/RhodeIsland.zip',
 'South Carolina': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/SouthCarolina.zip',
 'South Dakota': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/SouthDakota.zip',
 'Tennessee': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Tennessee.zip',
 'Texas': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Texas.zip',
 'Utah': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Utah.zip',
 'Vermont': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Vermont.zip',
 'Virginia': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Virginia.zip',
 'Washington': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Washington.zip',
 'West Virginia': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/WestVirginia.zip',
 'Wisconsin': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Wisconsin.zip',
 'Wyoming': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Wyoming.zip'}


from yacs.config import CfgNode as CN

# -----------------------------------------------------------------------------
# Config definition
# -----------------------------------------------------------------------------

_C = CN()
_C.DIR = "ckpt/ade20k-resnet50dilated-ppm_deepsup"

# -----------------------------------------------------------------------------
# Dataset
# -----------------------------------------------------------------------------
_C.DATASET = CN()
_C.DATASET.root_dataset = "./data/"
_C.DATASET.list_train = "./data/training.odgt"
_C.DATASET.list_val = "./data/validation.odgt"
_C.DATASET.num_class = 150
# multiscale train/test, size of short edge (int or tuple)
_C.DATASET.imgSizes = (300, 375, 450, 525, 600)
# maximum input image size of long edge
_C.DATASET.imgMaxSize = 1000
# maxmimum downsampling rate of the network
_C.DATASET.padding_constant = 8
# downsampling rate of the segmentation label
_C.DATASET.segm_downsampling_rate = 8
# randomly horizontally flip images when train/test
_C.DATASET.random_flip = True

# -----------------------------------------------------------------------------
# Model
# -----------------------------------------------------------------------------
_C.MODEL = CN()
# architecture of net_encoder
_C.MODEL.arch_encoder = "resnet50dilated"
# architecture of net_decoder
_C.MODEL.arch_decoder = "ppm_deepsup"
# weights to finetune net_encoder
_C.MODEL.weights_encoder = ""
# weights to finetune net_decoder
_C.MODEL.weights_decoder = ""
# number of feature channels between encoder and decoder
_C.MODEL.fc_dim = 2048

# -----------------------------------------------------------------------------
# Training
# -----------------------------------------------------------------------------
_C.TRAIN = CN()
_C.TRAIN.batch_size_per_gpu = 2
# epochs to train for
_C.TRAIN.num_epoch = 20
# epoch to start training. useful if continue from a checkpoint
_C.TRAIN.start_epoch = 0
# iterations of each epoch (irrelevant to batch size)
_C.TRAIN.epoch_iters = 5000

_C.TRAIN.optim = "SGD"
_C.TRAIN.lr_encoder = 0.02
_C.TRAIN.lr_decoder = 0.02
# power in poly to drop LR
_C.TRAIN.lr_pow = 0.9
# momentum for sgd, beta1 for adam
_C.TRAIN.beta1 = 0.9
# weights regularizer
_C.TRAIN.weight_decay = 1e-4
# the weighting of deep supervision loss
_C.TRAIN.deep_sup_scale = 0.4
# fix bn params, only under finetuning
_C.TRAIN.fix_bn = False
# number of data loading workers
_C.TRAIN.workers = 16

# frequency to display
_C.TRAIN.disp_iter = 20
# manual seed
_C.TRAIN.seed = 304

# -----------------------------------------------------------------------------
# Validation
# -----------------------------------------------------------------------------
_C.VAL = CN()
# currently only supports 1
_C.VAL.batch_size = 1
# output visualization during validation
_C.VAL.visualize = False
# the checkpoint to evaluate on
_C.VAL.checkpoint = "epoch_20.pth"

# -----------------------------------------------------------------------------
# Testing
# -----------------------------------------------------------------------------
_C.TEST = CN()
# currently only supports 1
_C.TEST.batch_size = 1
# the checkpoint to test on
_C.TEST.checkpoint = "epoch_20.pth"
# folder to output visualization results
_C.TEST.result = "./"



# System libs
import os
import argparse
from distutils.version import LooseVersion
# Numerical libs
import numpy as np
import torch
import torch.nn as nn
from scipy.io import loadmat
import csv
# Our libs
from dataset import TestDataset
from models import ModelBuilder, SegmentationModule
from utils import colorEncode, find_recursive, setup_logger
from lib.nn import user_scattered_collate, async_copy_to
from lib.utils import as_numpy
from PIL import Image
from tqdm import tqdm
from config import cfg

colors = loadmat('data/color150.mat')['colors']
names = {}
with open('data/object150_info.csv') as f:
    reader = csv.reader(f)
    next(reader)
    for row in reader:
        names[int(row[0])] = row[5].split(";")[0]


def visualize_result(data, pred, cfg):
    (img, info) = data

    # print predictions in descending order
    pred = np.int32(pred)
    pixs = pred.size
    uniques, counts = np.unique(pred, return_counts=True)
    print("Predictions in [{}]:".format(info))
    for idx in np.argsort(counts)[::-1]:
        name = names[uniques[idx] + 1]
        ratio = counts[idx] / pixs * 100
        if ratio > 0.1:
            print("  {}: {:.2f}%".format(name, ratio))

    # colorize prediction
    pred_color = colorEncode(pred, colors).astype(np.uint8)

    # aggregate images and save
    im_vis = np.concatenate((img, pred_color), axis=1)

    img_name = info.split('/')[-1]
    Image.fromarray(im_vis).save(
        os.path.join(cfg.TEST.result, img_name.replace('.jpg', '.png')))


def test(segmentation_module, loader, gpu):
    segmentation_module.eval()

    pbar = tqdm(total=len(loader))
    for batch_data in loader:
        # process data
        batch_data = batch_data[0]
        segSize = (batch_data['img_ori'].shape[0],
                   batch_data['img_ori'].shape[1])
        img_resized_list = batch_data['img_data']

        with torch.no_grad():
            scores = torch.zeros(1, cfg.DATASET.num_class, segSize[0], segSize[1])
            scores = async_copy_to(scores, gpu)

            for img in img_resized_list:
                feed_dict = batch_data.copy()
                feed_dict['img_data'] = img
                del feed_dict['img_ori']
                del feed_dict['info']
                feed_dict = \
                    (feed_dict, gpu)

                # forward pass
                pred_tmp = segmentation_module(feed_dict, segSize=segSize)
                scores = scores + pred_tmp / len(cfg.DATASET.imgSizes)

            _, pred = torch.max(scores, dim=1)
            pred = as_numpy(pred.squeeze(0).cpu())

        # visualization
        visualize_result(
            (batch_data['img_ori'], batch_data['info']),
            pred,
            cfg
        )

        pbar.update(1)


def main(cfg, gpu):
    torch.cuda.set_device(gpu)

    # Network Builders
    net_encoder = ModelBuilder.build_encoder(
        arch=cfg.MODEL.arch_encoder,
        fc_dim=cfg.MODEL.fc_dim,
        weights=cfg.MODEL.weights_encoder)
    net_decoder = ModelBuilder.build_decoder(
        arch=cfg.MODEL.arch_decoder,
        fc_dim=cfg.MODEL.fc_dim,
        num_class=cfg.DATASET.num_class,
        weights=cfg.MODEL.weights_decoder,
        use_softmax=True)

    crit = nn.NLLLoss(ignore_index=-1)

    segmentation_module = SegmentationModule(net_encoder, net_decoder, crit)

    # Dataset and Loader
    dataset_test = TestDataset(
        cfg.list_test,
        cfg.DATASET)
    loader_test = torch.utils.data.DataLoader(
        dataset_test,
        batch_size=cfg.TEST.batch_size,
        shuffle=False,
        collate_fn=user_scattered_collate,
        num_workers=5,
        drop_last=True)

    segmentation_module.cuda()

    # Main loop
    test(segmentation_module, loader_test, gpu)

    print('Inference done!')


if __name__ == '__main__':
    assert LooseVersion(torch.__version__) >= LooseVersion('0.4.0'), \
        'PyTorch>=0.4.0 is required'

    parser = argparse.ArgumentParser(
        description="PyTorch Semantic Segmentation Testing"
    )
    parser.add_argument(
        "--imgs",
        required=True,
        type=str,
        help="an image paths, or a directory name"
    )
    parser.add_argument(
        "--cfg",
        default="config/ade20k-resnet50dilated-ppm_deepsup.yaml",
        metavar="FILE",
        help="path to config file",
        type=str,
    )
    parser.add_argument(
        "--gpu",
        default=0,
        type=int,
        help="gpu id for evaluation"
    )
    parser.add_argument(
        "opts",
        help="Modify config options using the command-line",
        default=None,
        nargs=argparse.REMAINDER,
    )
    args = parser.parse_args()

    cfg.merge_from_file(args.cfg)
    cfg.merge_from_list(args.opts)
    # cfg.freeze()

    logger = setup_logger(distributed_rank=0)   # TODO
    logger.info("Loaded configuration file {}".format(args.cfg))
    logger.info("Running with config:\n{}".format(cfg))

    cfg.MODEL.arch_encoder = cfg.MODEL.arch_encoder.lower()
    cfg.MODEL.arch_decoder = cfg.MODEL.arch_decoder.lower()

    # absolute paths of model weights
    cfg.MODEL.weights_encoder = os.path.join(
        cfg.DIR, 'encoder_' + cfg.TEST.checkpoint)
    cfg.MODEL.weights_decoder = os.path.join(
        cfg.DIR, 'decoder_' + cfg.TEST.checkpoint)

    assert os.path.exists(cfg.MODEL.weights_encoder) and \
        os.path.exists(cfg.MODEL.weights_decoder), "checkpoint does not exitst!"

    # generate testing image list
    if os.path.isdir(args.imgs[0]):
        imgs = find_recursive(args.imgs[0])
    else:
        imgs = [args.imgs]
    assert len(imgs), "imgs should be a path to image (.jpg) or directory."
    cfg.list_test = [{'fpath_img': x} for x in imgs]

    if not os.path.isdir(cfg.TEST.result):
        os.makedirs(cfg.TEST.result)

    main(cfg, args.gpu)



import math
import torch
from torch.optim.optimizer import Optimizer, required


class RAdam(Optimizer):

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, degenerated_to_sgd=True):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))

        self.degenerated_to_sgd = degenerated_to_sgd
        if isinstance(params, (list, tuple)) and len(params) > 0 and isinstance(params[0], dict):
            for param in params:
                if 'betas' in param and (param['betas'][0] != betas[0] or param['betas'][1] != betas[1]):
                    param['buffer'] = [[None, None, None] for _ in range(10)]
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay,
                        buffer=[[None, None, None] for _ in range(10)])
        super(RAdam, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(RAdam, self).__setstate__(state)

    def step(self, closure=None):

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError('RAdam does not support sparse gradients')

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)
                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                exp_avg.mul_(beta1).add_(1 - beta1, grad)

                state['step'] += 1
                buffered = group['buffer'][int(state['step'] % 10)]
                if state['step'] == buffered[0]:
                    N_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state['step']
                    beta2_t = beta2 ** state['step']
                    N_sma_max = 2 / (1 - beta2) - 1
                    N_sma = N_sma_max - 2 * state['step'] * beta2_t / (1 - beta2_t)
                    buffered[1] = N_sma

                    # more conservative since it's an approximated value
                    if N_sma >= 5:
                        step_size = math.sqrt(
                            (1 - beta2_t) * (N_sma - 4) / (N_sma_max - 4) * (N_sma - 2) / N_sma * N_sma_max / (
                                        N_sma_max - 2)) / (1 - beta1 ** state['step'])
                    elif self.degenerated_to_sgd:
                        step_size = 1.0 / (1 - beta1 ** state['step'])
                    else:
                        step_size = -1
                    buffered[2] = step_size

                # more conservative since it's an approximated value
                if N_sma >= 5:
                    if group['weight_decay'] != 0:
                        p_data_fp32.add_(-group['weight_decay'] * group['lr'], p_data_fp32)
                    denom = exp_avg_sq.sqrt().add_(group['eps'])
                    p_data_fp32.addcdiv_(-step_size * group['lr'], exp_avg, denom)
                    p.data.copy_(p_data_fp32)
                elif step_size > 0:
                    if group['weight_decay'] != 0:
                        p_data_fp32.add_(-group['weight_decay'] * group['lr'], p_data_fp32)
                    p_data_fp32.add_(-step_size * group['lr'], exp_avg)
                    p.data.copy_(p_data_fp32)

        return loss


class PlainRAdam(Optimizer):

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, degenerated_to_sgd=True):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))

        self.degenerated_to_sgd = degenerated_to_sgd
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)

        super(PlainRAdam, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(PlainRAdam, self).__setstate__(state)

    def step(self, closure=None):

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError('RAdam does not support sparse gradients')

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)
                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                exp_avg.mul_(beta1).add_(1 - beta1, grad)

                state['step'] += 1
                beta2_t = beta2 ** state['step']
                N_sma_max = 2 / (1 - beta2) - 1
                N_sma = N_sma_max - 2 * state['step'] * beta2_t / (1 - beta2_t)

                # more conservative since it's an approximated value
                if N_sma >= 5:
                    if group['weight_decay'] != 0:
                        p_data_fp32.add_(-group['weight_decay'] * group['lr'], p_data_fp32)
                    step_size = group['lr'] * math.sqrt(
                        (1 - beta2_t) * (N_sma - 4) / (N_sma_max - 4) * (N_sma - 2) / N_sma * N_sma_max / (
                                    N_sma_max - 2)) / (1 - beta1 ** state['step'])
                    denom = exp_avg_sq.sqrt().add_(group['eps'])
                    p_data_fp32.addcdiv_(-step_size, exp_avg, denom)
                    p.data.copy_(p_data_fp32)
                elif self.degenerated_to_sgd:
                    if group['weight_decay'] != 0:
                        p_data_fp32.add_(-group['weight_decay'] * group['lr'], p_data_fp32)
                    step_size = group['lr'] / (1 - beta1 ** state['step'])
                    p_data_fp32.add_(-step_size, exp_avg)
                    p.data.copy_(p_data_fp32)

        return loss


class AdamW(Optimizer):

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, warmup=0):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))

        defaults = dict(lr=lr, betas=betas, eps=eps,
                        weight_decay=weight_decay, warmup=warmup)
        super(AdamW, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(AdamW, self).__setstate__(state)

    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError('Adam does not support sparse gradients, please consider SparseAdam instead')

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)
                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1

                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                exp_avg.mul_(beta1).add_(1 - beta1, grad)

                denom = exp_avg_sq.sqrt().add_(group['eps'])
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                if group['warmup'] > state['step']:
                    scheduled_lr = 1e-8 + state['step'] * group['lr'] / group['warmup']
                else:
                    scheduled_lr = group['lr']

                step_size = scheduled_lr * math.sqrt(bias_correction2) / bias_correction1

                if group['weight_decay'] != 0:
                    p_data_fp32.add_(-group['weight_decay'] * scheduled_lr, p_data_fp32)

                p_data_fp32.addcdiv_(-step_size, exp_avg, denom)

                p.data.copy_(p_data_fp32)

        return loss



import torch
import torch.nn as nn

class ResidualBlock1D(nn.Module):
    def __init__(self, in_channel, out_channel, stride=1):
        super(ResidualBlock1D, self).__init__()
        self.in_channel = in_channel
        self.out_channel = out_channel
        self.stride = stride
        self.bn1 = nn.BatchNorm1d(in_channel)
        self.relu = nn.ReLU(inplace=True)
        self.conv1 = nn.Conv1d(in_channel, int(out_channel/4), 1, 1, bias=False)
        self.bn2 = nn.BatchNorm1d(int(out_channel/4))
        self.conv2 = nn.Conv1d(int(out_channel / 4), int(out_channel / 4), 3, stride=stride, padding = 1, bias=False)
        self.bn3 = nn.BatchNorm1d(int(out_channel/4))
        self.conv3 = nn.Conv1d(int(out_channel / 4), out_channel, 1, stride=1, bias=False)
        self.conv4 = nn.Conv1d(in_channel, out_channel, 1, stride=stride, bias=False)
    def forward(self, x):
        res = x
        x = self.bn1(x)
        out_1 = self.relu(x)
        x = out_1
        x = self.conv1(x)
        x = self.bn2(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.bn3(x)
        x = self.relu(x)
        x = self.conv3(x)
        if(self.in_channel != self.out_channel) or (self.stride!=1):
            res = self.conv4(out_1)
        x += res
        return x


class ResidualBlock(nn.Module):
    def __init__(self, in_channel, out_channel, stride=1):
        super(ResidualBlock, self).__init__()
        self.in_channel = in_channel
        self.out_channel = out_channel
        self.stride = stride
        self.bn1 = nn.BatchNorm2d(in_channel)
        self.relu = nn.ReLU(inplace=True)
        self.conv1 = nn.Conv2d(in_channel, int(out_channel/4), 1, 1, bias=False)
        self.bn2 = nn.BatchNorm2d(int(out_channel/4))
        self.conv2 = nn.Conv2d(int(out_channel / 4), int(out_channel / 4), 3, stride=stride, padding = 1, bias=False)
        self.bn3 = nn.BatchNorm2d(int(out_channel/4))
        self.conv3 = nn.Conv2d(int(out_channel / 4), out_channel, 1, stride=1, bias=False)
        self.conv4 = nn.Conv2d(in_channel, out_channel, 1, stride=stride, bias=False)
    def forward(self, x):
        res = x
        x = self.bn1(x)
        out_1 = self.relu(x)
        x = out_1
        x = self.conv1(x)
        x = self.bn2(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.bn3(x)
        x = self.relu(x)
        x = self.conv3(x)
        if(self.in_channel != self.out_channel) or (self.stride!=1):
            res = self.conv4(out_1)
        x += res
        return x



class ResidualBlock(nn.Module):
    def __init__(self, in_channel, out_channel, stride=1):
        super(ResidualBlock, self).__init__()
        self.in_channel = in_channel
        self.out_channel = out_channel
        self.stride = stride
        self.bn1 = nn.BatchNorm2d(in_channel)
        self.relu = nn.ReLU(inplace=True)
        self.conv1 = nn.Conv2d(in_channel, int(out_channel/4), 1, 1, bias=False)
        self.bn2 = nn.BatchNorm2d(int(out_channel/4))
        self.conv2 = nn.Conv2d(int(out_channel / 4), int(out_channel / 4), 3, stride=stride, padding = 1, bias=False)
        self.bn3 = nn.BatchNorm2d(int(out_channel/4))
        self.conv3 = nn.Conv2d(int(out_channel / 4), out_channel, 1, stride=1, bias=False)
        self.conv4 = nn.Conv2d(in_channel, out_channel, 1, stride=stride, bias=False)
    def forward(self, x):
        res = x
        x = self.bn1(x)
        out_1 = self.relu(x)
        x = out_1
        x = self.conv1(x)
        x = self.bn2(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.bn3(x)
        x = self.relu(x)
        x = self.conv3(x)
        if(self.in_channel != self.out_channel) or (self.stride!=1):
            res = self.conv4(out_1)
        x += res
        return x




import json 
import xmltodict 
import glob
import os
import subprocess
import numpy as np
import xml.etree.ElementTree as ET
import pandas as pd

def processcvat(xmlfiles, namemap):

    imgPath = []
    xtl = []
    ytl = []
    xbr = []
    ybr = []
    use = []
    security = []
    material = []
    condition = []
    completeness = []
    width = []
    height = []
    
    for f in xmlfiles:
        
        with open(f) as xml_file:
            fstr = xml_file.read()
            
            if namemap:
                for k in namemap:
                    fstr = fstr.replace(k, namemap[k])
            
            j = xmltodict.parse(fstr)
            
            imgs = j['annotations']['image']
            for img in imgs:
                name = img['@name']
                w = img['@width']
                h = img['@height']
                
                try:
                    bs = img['box']

                    for b in bs:
                            
                        xtl.append(b['@xtl'])
                        ytl.append(b['@ytl'])
                        xbr.append(b['@xbr'])
                        ybr.append(b['@ybr'])
                        attributes = b['attribute']
                        imgPath.append(name)
                        width.append(w)
                        height.append(h)
                        thisatt = {}
                        for att in attributes:
                            attName = att['@name']
                            attValue = att['#text']
                            thisatt[attName] = attValue
                        use.append(thisatt['use'])
                        security.append(thisatt['security'])
                        material.append(thisatt['material'])
                        condition.append(thisatt['condition'])
                        completeness.append(thisatt['completeness'])

                except: pass
                
                try:
                    
                    xtl.append(bs['@xtl'])
                    ytl.append(bs['@ytl'])
                    xbr.append(bs['@xbr'])
                    ybr.append(bs['@ybr'])


                    attributes = bs['attribute']
                    imgPath.append(name)
                    width.append(w)
                    height.append(h)
                    thisatt = {}
                    for att in attributes:
                        attName = att['@name']
                        attValue = att['#text']
                        thisatt[attName] = attValue
                    use.append(thisatt['use'])
                    security.append(thisatt['security'])
                    material.append(thisatt['material'])
                    condition.append(thisatt['condition'])
                    completeness.append(thisatt['completeness'])

                except: pass
                
          
    df = pd.DataFrame(list(zip(imgPath,width,height,xtl,ytl,xbr,ybr, material,use,condition,security,completeness)), columns =['imgPath','width','height','xtl','ytl','xbr','ybr', 'material','use','condition','security','completeness'])

    return df



import torch
import torchvision.transforms as transforms
import numpy as np
import sklearn.covariance
from PIL import Image
from npid_models import resnet
import pandas as pd
import torch.nn.functional as F
from torch.autograd.variable import Variable



def resize2d(img, size):
    with torch.no_grad():
        return (F.adaptive_avg_pool2d(Variable(img), size)).data


class Distance_Eval(object):
    
    def __init__(self, checkpoint, low_dim=128):
        
        self.image_decorator = None
        model = resnet.__dict__['resnet50'](pretrained=False,low_dim=low_dim)
        model = torch.nn.DataParallel(model).cuda()
        checkpoint = torch.load(checkpoint)
        model.load_state_dict(checkpoint['state_dict'])

        lemniscate = checkpoint['lemniscate']            
        
        calc = True
    
        self.net = model
    
        self.net.eval()
        trainFeatures = lemniscate.memory.t()
        trainFeatures = trainFeatures.cpu().numpy()
        
        normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                 std=[0.229, 0.224, 0.225])
        
        self.transform = transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            normalize
        ])
    
        if calc == True:
            self.mu, self.precision = self._fitGaussian(trainFeatures)
            np.save("precision", self.precision)
            np.save("mu", self.mu)
        else:
            self.mu = np.load("mu.npy")
            self.precision = np.load("precision.npy")
    
    def _fitGaussian(self, features):
        mu = np.mean(features, axis=1)
    
        cov = sklearn.covariance.EmpiricalCovariance(store_precision=True, assume_centered=False)
        cov.fit(features.transpose())
        prec = cov.precision_
    
        return mu, prec
    
    def _mahalanobis(self, x, mu, precision):
        x = x.squeeze()
        distance = np.matmul(x - mu, np.matmul(precision, x - mu))
        return distance
    
    def get_distance_and_features(self, pil_image):
  
        if self.transform is not None:
            # resize to 224
            pil_image.convert('RGB')
            pil_image = self.transform(pil_image)            
        
        with torch.no_grad():
            img = torch.FloatTensor(pil_image).unsqueeze(0)            
            # img = resize2d(img,(224,224))
            features = self.net(img)
            features = features.cpu().numpy()
            
        return self._mahalanobis(features, self.mu, self.precision), features
    



if __name__ == '__main__':
    print ("Test")
    # test_path = '/data/PixelWise/BDD10k_clear_daytime/train/train/4c840aee-0dd28451.jpg'
    # test_path = '/data/PixelWise/BDD10k_clear_daytime/train/train/7171644d-4e74752e.jpg'
    # test_path = '/data/PixelWise/BDD10k_night/train/train/2a051b12-a7404163.jpg'
    # test_path = '/data/PixelWise/BDD10k_clear_daytime/train/train/9e86ec05-56319156.jpg'
    # test_path = '/data/PixelWise/BDD10k_clear_daytime/train/train/851ec515-15faca1d.jpg'
    # test_path = '/data/PixelWise/BDD10k_dusk/train/train/2a2d54fa-fd3fc827.jpg'
    # test_path = sys.argv[1]
    
    # Checkpoint no pedestrians
    evaluator = Distance_Eval('/data/PixelWise/NoPed_BDD100k_e100ch.pth.tar')
    
    # Checkpoint day night
    # evaluator = Distance_Eval('/data/PixelWise/cdt10k_e100ch.pth.tar')
    
    test_path = '/data/PixelWise/BDD100k_NoPed/WithPed/769136f0-17e25e15.jpg'
    evaluator.image_check(test_path)



import torchvision.datasets as datasets
import os
import numpy as np
from PIL import Image
from torch.utils.data import Dataset
import pandas as pd
import torch

class ImageFolderInstance(datasets.ImageFolder):
    """: Folder datasets which returns the index of the image as well::
    """
    def __getitem__(self, index):
        """
        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is class_index of the target class.
        """
        path, target = self.imgs[index]
        img = self.loader(path)
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target, index

class MaskFolderInstance(datasets.ImageFolder):
    """: Slightly hackish way to train on masks quick
    """
    def __init__(self, *args, **kwargs):
        if 'mask_images' in kwargs.keys():
            self.mask_images = kwargs['mask_images']
            kwargs.pop('mask_images')
        else:
            self.mask_images = False
        super().__init__(*args, **kwargs)
        # Another hackish thing to get rid of a problem with missin masks.
        # This class has to be redesigned from scratch

        # The following lines can be used to remove images without mask. This is only relevant
        # when doing the split
        # no_mask_ids = []
        # for i,entry in enumerate(self.imgs):
        #    if not os.path.isfile('/home/saschaho/Simcenter/Floor_Elevation_Data/Streetview_Irma/Streetview_Irma/images/' + os.path.basename(entry[0]).replace('.jpg','.png')):
        #        no_mask_ids.append(i)
        # # reverse order will preserver order during deletion
        # for i in sorted(no_mask_ids, reverse=True):
        #     del self.imgs[i]


    def __getitem__(self, index):
        """
        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is class_index of the target class.
        """
        path, target = self.imgs[index]

        img = self.loader(path)

        if self.mask_images:
            # replace jpg image with mask which is saved somewhere else.
            # This implementation needs to be more generalized
            #mask_filename = '/home/saschaho/Simcenter/Floor_Elevation_Data/Streetview_Irma/Streetview_Irma/images/' + os.path.basename(path).replace('.jpg','.png')
            mask_filename = os.path.basename(path).replace('.jpg','.png')
            mask = Image.open(mask_filename)
            mask = np.array(mask)
            # Filter building labels
            mask[np.where((mask != 25) & (mask != 1))] = 0
            img = np.array(img)
            img[mask == 0, :] = 0
            img = Image.fromarray(np.uint8(img))

        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target, index


class CombinedMaskDataset(Dataset):
    """: Slightly hackish way to train on masks quick
    """
    def __init__(self, other_data_path=None, csv_root_folder=None, data_csv=None, transform=None, mask_images=True, attribute=None):
        self.mask_images = mask_images
        self.transform = transform
        self.imgs = []
        self.targets = []
        self.target_transform = None

        # Add only training images from the training data
        train_data_pd = pd.read_csv(data_csv)
        for entry in train_data_pd.iterrows():
            jpg_file = os.path.join(csv_root_folder, entry[1]['filename'])
            # If image and mask file exists. Add to list
            if os.path.isfile(jpg_file) and os.path.isfile(os.path.join(csv_root_folder, entry[1]['filename'].replace('.jpg', '.png'))):
                self.imgs.append(jpg_file)
                self.targets.append(entry[1][attribute])

        if other_data_path is not None:
            # If this is training data add all images from the other data
            for subdirs,_,filenames in os.walk(other_data_path,followlinks=True):
                for filename in filenames:
                    if 'jpg' in filename:
                        self.imgs.append(os.path.join(subdirs,filename))
                        self.targets.append(-1)

        self.train_labels = self.targets


    def __len__(self):
        return len(self.imgs)

    def loader(self,path):
        with open(path, 'rb') as f:
            img = Image.open(f)
            return img.convert('RGB')

    def __getitem__(self, index):
        """
        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is class_index of the target class.
        """
        path= self.imgs[index]
        target = torch.FloatTensor([float(self.targets[index])])

        img = self.loader(path)

        if self.mask_images:
            # replace jpg image with mask which is saved somewhere else.
            # This implementation needs to be more generalized
            #mask_filename = '/home/saschaho/Simcenter/Floor_Elevation_Data/Streetview_Irma/Streetview_Irma/images/' + os.path.basename(path).replace('.jpg','.png')
            mask_filename = path.replace('.jpg','.png')
            mask = Image.open(mask_filename)
            mask = np.array(mask)
            # Filter building labels
            mask[np.where((mask != 25) & (mask != 1))] = 0
            img = np.array(img)
            img[mask == 0, :] = 0
            img = Image.fromarray(np.uint8(img))

        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target, index


import torch.nn as nn
import torch
    

class PMG(nn.Module):
    def __init__(self, model, feature_size, classes_num):
        super(PMG, self).__init__()

        self.features = model
        self.max1 = nn.MaxPool2d(kernel_size=56, stride=56)
        self.max2 = nn.MaxPool2d(kernel_size=28, stride=28)
        self.max3 = nn.MaxPool2d(kernel_size=14, stride=14)
        self.num_ftrs = 2048 * 1 * 1
        self.elu = nn.ELU(inplace=True)

        self.classifier_concat = nn.Sequential(
            nn.BatchNorm1d(1024 * 3),
            nn.Linear(1024 * 3, feature_size),
            nn.BatchNorm1d(feature_size),
            nn.ELU(inplace=True),
            nn.Linear(feature_size, classes_num),
        )

        self.conv_block1 = nn.Sequential(
            BasicConv(self.num_ftrs//4, feature_size, kernel_size=1, stride=1, padding=0, relu=True),
            BasicConv(feature_size, self.num_ftrs//2, kernel_size=3, stride=1, padding=1, relu=True)
        )
        self.classifier1 = nn.Sequential(
            nn.BatchNorm1d(self.num_ftrs//2),
            nn.Linear(self.num_ftrs//2, feature_size),
            nn.BatchNorm1d(feature_size),
            nn.ELU(inplace=True),
            nn.Linear(feature_size, classes_num),
        )

        self.conv_block2 = nn.Sequential(
            BasicConv(self.num_ftrs//2, feature_size, kernel_size=1, stride=1, padding=0, relu=True),
            BasicConv(feature_size, self.num_ftrs//2, kernel_size=3, stride=1, padding=1, relu=True)
        )
        self.classifier2 = nn.Sequential(
            nn.BatchNorm1d(self.num_ftrs//2),
            nn.Linear(self.num_ftrs//2, feature_size),
            nn.BatchNorm1d(feature_size),
            nn.ELU(inplace=True),
            nn.Linear(feature_size, classes_num),
        )

        self.conv_block3 = nn.Sequential(
            BasicConv(self.num_ftrs, feature_size, kernel_size=1, stride=1, padding=0, relu=True),
            BasicConv(feature_size, self.num_ftrs//2, kernel_size=3, stride=1, padding=1, relu=True)
        )
        self.classifier3 = nn.Sequential(
            nn.BatchNorm1d(self.num_ftrs//2),
            nn.Linear(self.num_ftrs//2, feature_size),
            nn.BatchNorm1d(feature_size),
            nn.ELU(inplace=True),
            nn.Linear(feature_size, classes_num),
        )

    def forward(self, x):
        xf1, xf2, xf3, xf4, xf5 = self.features(x)

        xl1 = self.conv_block1(xf3)
        xl2 = self.conv_block2(xf4)
        xl3 = self.conv_block3(xf5)
        
        xl1 = self.max1(xl1)
        xl1 = xl1.view(xl1.size(0), -1)
        xc1 = self.classifier1(xl1)

        xl2 = self.max2(xl2)
        xl2 = xl2.view(xl2.size(0), -1)
        xc2 = self.classifier2(xl2)

        xl3 = self.max3(xl3)
        xl3 = xl3.view(xl3.size(0), -1)
        xc3 = self.classifier3(xl3)
          
        x_concat = torch.cat((xl1, xl2, xl3), -1)
        x_concat = self.classifier_concat(x_concat)
        return xc1, xc2, xc3, x_concat
    
    
class BasicConv(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, stride=1, padding=0, dilation=1, groups=1, relu=True, bn=True, bias=False):
        super(BasicConv, self).__init__()
        self.out_channels = out_planes
        self.conv = nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size,
                              stride=stride, padding=padding, dilation=dilation, groups=groups, bias=bias)
        self.bn = nn.BatchNorm2d(out_planes, eps=1e-5,
                                 momentum=0.01, affine=True) if bn else None
        self.relu = nn.ReLU() if relu else None

    def forward(self, x):
        x = self.conv(x)
        if self.bn is not None:
            x = self.bn(x)
        if self.relu is not None:
            x = self.relu(x)
        return x



from __future__ import print_function
from PIL import Image
import torchvision.datasets as datasets
import torch.utils.data as data

class CIFAR10Instance(datasets.CIFAR10):
    """CIFAR10Instance Dataset.
    """
    def __getitem__(self, index):
        if self.train:
            img, target = self.train_data[index], self.train_labels[index]
        else:
            img, target = self.test_data[index], self.test_labels[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target, index

class CIFAR100Instance(CIFAR10Instance):
    """CIFAR100Instance Dataset.

    This is a subclass of the `CIFAR10Instance` Dataset.
    """
    base_folder = 'cifar-100-python'
    url = "https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz"
    filename = "cifar-100-python.tar.gz"
    tgz_md5 = 'eb9058c3a382ffc7106e4002c42a8d85'
    train_list = [
        ['train', '16019d7e3df5f24257cddd939b257f8d'],
    ]

    test_list = [
        ['test', 'f0ef6b0ae62326f3e7ffdfab6717acfc'],
    ]



# Modified by bacetiner
import datetime
import os
import traceback

import torch
from torch import nn
from torch.utils.data import DataLoader
from torchvision import transforms
from lib.efficientdet.dataset import CocoDataset, Resizer, Normalizer, Augmenter, collater
from lib.backbone import EfficientDetBackbone
from tensorboardX import SummaryWriter
import numpy as np
from tqdm import tqdm

from lib.efficientdet.loss import FocalLoss
from lib.utils.sync_batchnorm import patch_replication_callback
from lib.utils.utils import replace_w_sync_bn, CustomDataParallel, get_last_weights, init_weights


import urllib

def boolean_string(s):
    if s not in {'False', 'True'}:
        raise ValueError('Not a valid boolean string')
    return s == 'True'


class ModelWithLoss(nn.Module):
    def __init__(self, model, debug=False):
        super().__init__()
        self.criterion = FocalLoss()
        self.model = model
        self.debug = debug

    def forward(self, imgs, annotations, obj_list=None):
        _, regression, classification, anchors = self.model(imgs)
        if self.debug:
            cls_loss, reg_loss = self.criterion(classification, regression, anchors, annotations,
                                                imgs=imgs, obj_list=obj_list)
        else:
            cls_loss, reg_loss = self.criterion(classification, regression, anchors, annotations)
        return cls_loss, reg_loss




class Detector():
    def __init__(self, verbose=1):
        self.system_dict = {};
        self.system_dict["verbose"] = verbose;
        self.system_dict["local"] = {};
        self.system_dict["dataset"] = {};
        self.system_dict["dataset"]["train"] = {};
        self.system_dict["dataset"]["val"] = {};
        self.system_dict["dataset"]["val"]["status"] = False;
        self.system_dict["params"] = {};

        self.set_fixed_params();

    
    def set_fixed_params(self):
        self.system_dict["params"]["project"] = "custom";
        self.system_dict["params"]["project_name"] = "custom";
        self.system_dict["params"]["mean"] = [0.485, 0.456, 0.406];
        self.system_dict["params"]["std"] = [0.229, 0.224, 0.225];
        self.system_dict["params"]["anchors_scales"] = '[2 ** 0, 2 ** (1.0 / 3.0), 2 ** (2.0 / 3.0)]';
        self.system_dict["params"]["anchors_ratios"] = '[(1.0, 1.0), (1.4, 0.7), (0.7, 1.4)]';
        self.system_dict["params"]["log_path"] = "logs/";
        self.system_dict["params"]["saved_path"] = "trained_weights/";
        self.system_dict["params"]["debug"] = False;


    def set_train_dataset(self, root_dir, coco_dir, img_dir, set_dir,  classes_list=[], batch_size=2, num_workers=3):
        self.system_dict["dataset"]["train"]["root_dir"] = root_dir
        self.system_dict["dataset"]["train"]["coco_dir"] = coco_dir
        self.system_dict["dataset"]["train"]["img_dir"] = img_dir
        self.system_dict["dataset"]["train"]["set_dir"] = set_dir

        self.system_dict["params"]["batch_size"] = batch_size;
        self.system_dict["params"]["num_workers"] = num_workers;
        self.system_dict["params"]["obj_list"] = classes_list;


    def set_val_dataset(self, root_dir, coco_dir, img_dir, set_dir):
        self.system_dict["dataset"]["val"]["root_dir"] = root_dir
        self.system_dict["dataset"]["val"]["coco_dir"] = coco_dir
        self.system_dict["dataset"]["val"]["img_dir"] = img_dir
        self.system_dict["dataset"]["val"]["set_dir"] = set_dir

        self.system_dict["dataset"]["val"]["status"] = True;


    #"efficientdet-d0.pth";
    #"efficientdet-d1.pth";
    #"efficientdet-d2.pth";
    #"efficientdet-d3.pth";
    #"efficientdet-d4.pth";
    #"efficientdet-d5.pth";
    #"efficientdet-d6.pth";
    #"efficientdet-d7.pth";
    def set_model(self, model_name="efficientdet-d0.pth", num_gpus=1, freeze_head=False):
        if("0" in model_name):
            self.system_dict["params"]["compound_coef"] = 0;
            self.system_dict["params"]["load_weights"] = "pretrained_weights/efficientdet-d0.pth";
        elif("1" in model_name):
            self.system_dict["params"]["compound_coef"] = 1;
            self.system_dict["params"]["load_weights"] = "pretrained_weights/efficientdet-d1.pth";
        elif("2" in model_name):
            self.system_dict["params"]["compound_coef"] = 2;
            self.system_dict["params"]["load_weights"] = "pretrained_weights/efficientdet-d2.pth";
        elif("3" in model_name):
            self.system_dict["params"]["compound_coef"] = 3;
            self.system_dict["params"]["load_weights"] = "pretrained_weights/efficientdet-d3.pth";
        elif("4" in model_name):
            self.system_dict["params"]["compound_coef"] = 4;
            self.system_dict["params"]["load_weights"] = "pretrained_weights/efficientdet-d4.pth";
        elif("5" in model_name):
            self.system_dict["params"]["compound_coef"] = 5;
            self.system_dict["params"]["load_weights"] = "pretrained_weights/efficientdet-d5.pth";
        elif("6" in model_name):
            self.system_dict["params"]["compound_coef"] = 6;
            self.system_dict["params"]["load_weights"] = "pretrained_weights/efficientdet-d6.pth";
        elif("7" in model_name):
            self.system_dict["params"]["compound_coef"] = 7;
            self.system_dict["params"]["load_weights"] = "pretrained_weights/efficientdet-d7.pth";

        self.system_dict["params"]["num_gpus"] = num_gpus;
        self.system_dict["params"]["head_only"] = freeze_head;
        

    #adamw
    #sgd
    def set_hyperparams(self, optimizer="adamw", lr=0.001, es_min_delta=0.0, es_patience=0):
        self.system_dict["params"]["lr"] = lr;
        self.system_dict["params"]["optim"] = optimizer; 
        self.system_dict["params"]["es_min_delta"] = es_min_delta;
        self.system_dict["params"]["es_patience"] = es_patience;


    def train(self, num_epochs=2, val_interval=1, save_interval=1):
        self.system_dict["params"]["num_epochs"] = num_epochs;
        self.system_dict["params"]["val_interval"] = val_interval;
        self.system_dict["params"]["save_interval"] = save_interval;

        self.start_training();


    def start_training(self):
        if self.system_dict["params"]["num_gpus"] == 0:
            os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

        if torch.cuda.is_available():
            torch.cuda.manual_seed(42)
        else:
            torch.manual_seed(42)

        self.system_dict["params"]["saved_path"] = self.system_dict["params"]["saved_path"] + "/" + self.system_dict["params"]["project_name"] + "/";
        self.system_dict["params"]["log_path"] = self.system_dict["params"]["log_path"] + "/" + self.system_dict["params"]["project_name"] + "/tensorboard/";
        os.makedirs(self.system_dict["params"]["saved_path"], exist_ok=True)
        os.makedirs(self.system_dict["params"]["log_path"], exist_ok=True)

        training_params = {'batch_size': self.system_dict["params"]["batch_size"],
                   'shuffle': True,
                   'drop_last': True,
                   'collate_fn': collater,
                   'num_workers': self.system_dict["params"]["num_workers"]}

        val_params = {'batch_size': self.system_dict["params"]["batch_size"],
                      'shuffle': False,
                      'drop_last': True,
                      'collate_fn': collater,
                      'num_workers': self.system_dict["params"]["num_workers"]}

        input_sizes = [512, 640, 768, 896, 1024, 1280, 1280, 1536]
        training_set = CocoDataset(self.system_dict["dataset"]["train"]["root_dir"], 
                                    self.system_dict["dataset"]["train"]["coco_dir"], 
                                    self.system_dict["dataset"]["train"]["img_dir"], 
                                    set_dir=self.system_dict["dataset"]["train"]["set_dir"],
                                    transform=transforms.Compose([Normalizer(mean=self.system_dict["params"]["mean"], std=self.system_dict["params"]["std"]),
                                                                 Augmenter(),
                                                                 Resizer(input_sizes[self.system_dict["params"]["compound_coef"]])]))
        training_generator = DataLoader(training_set, **training_params)

        if(self.system_dict["dataset"]["val"]["status"]):
            val_set = CocoDataset(self.system_dict["dataset"]["val"]["root_dir"], 
                                    self.system_dict["dataset"]["val"]["coco_dir"], 
                                    self.system_dict["dataset"]["val"]["img_dir"], 
                                    set_dir=self.system_dict["dataset"]["val"]["set_dir"],
                                    transform=transforms.Compose([Normalizer(self.system_dict["params"]["mean"], self.system_dict["params"]["std"]),
                                                             Resizer(input_sizes[self.system_dict["params"]["compound_coef"]])]))
            val_generator = DataLoader(val_set, **val_params)

        print("");
        print("");
        model = EfficientDetBackbone(num_classes=len(self.system_dict["params"]["obj_list"]), 
                                        compound_coef=self.system_dict["params"]["compound_coef"],
                                        ratios=eval(self.system_dict["params"]["anchors_ratios"]), 
                                        scales=eval(self.system_dict["params"]["anchors_scales"]));

        os.makedirs("pretrained_weights", exist_ok=True);

        if(self.system_dict["params"]["compound_coef"] == 0):
            if(not os.path.isfile(self.system_dict["params"]["load_weights"])):
                print("Downloading weights");
                urllib.request.urlretrieve("https://github.com/zylo117/Yet-Another-Efficient-Pytorch/releases/download/1.0/efficientdet-d0.pth", filename=self.system_dict["params"]["load_weights"])
        elif(self.system_dict["params"]["compound_coef"] == 1):
            if(not os.path.isfile(self.system_dict["params"]["load_weights"])):
                print("Downloading weights");
                urllib.request.urlretrieve("https://github.com/zylo117/Yet-Another-Efficient-Pytorch/releases/download/1.0/efficientdet-d1.pth", filename=self.system_dict["params"]["load_weights"])
        elif(self.system_dict["params"]["compound_coef"] == 2):
            if(not os.path.isfile(self.system_dict["params"]["load_weights"])):
                print("Downloading weights");
                urllib.request.urlretrieve("https://github.com/zylo117/Yet-Another-Efficient-Pytorch/releases/download/1.0/efficientdet-d2.pth", filename=self.system_dict["params"]["load_weights"])
        elif(self.system_dict["params"]["compound_coef"] == 3):
            if(not os.path.isfile(self.system_dict["params"]["load_weights"])):
                print("Downloading weights");
                urllib.request.urlretrieve("https://github.com/zylo117/Yet-Another-Efficient-Pytorch/releases/download/1.0/efficientdet-d3.pth", filename=self.system_dict["params"]["load_weights"])
        elif(self.system_dict["params"]["compound_coef"] == 4):
            if(not os.path.isfile(self.system_dict["params"]["load_weights"])):
                print("Downloading weights");
                urllib.request.urlretrieve("https://github.com/zylo117/Yet-Another-Efficient-Pytorch/releases/download/1.0/efficientdet-d4.pth", filename=self.system_dict["params"]["load_weights"])
        elif(self.system_dict["params"]["compound_coef"] == 5):
            if(not os.path.isfile(self.system_dict["params"]["load_weights"])):
                print("Downloading weights");
                urllib.request.urlretrieve("https://github.com/zylo117/Yet-Another-Efficient-Pytorch/releases/download/1.0/efficientdet-d5.pth", filename=self.system_dict["params"]["load_weights"])
        elif(self.system_dict["params"]["compound_coef"] == 6):
            if(not os.path.isfile(self.system_dict["params"]["load_weights"])):
                print("Downloading weights");
                urllib.request.urlretrieve("https://github.com/zylo117/Yet-Another-Efficient-Pytorch/releases/download/1.0/efficientdet-d6.pth", filename=self.system_dict["params"]["load_weights"])
        elif(self.system_dict["params"]["compound_coef"] == 7):
            if(not os.path.isfile(self.system_dict["params"]["load_weights"])):
                print("Downloading weights");
                urllib.request.urlretrieve("https://github.com/zylo117/Yet-Another-Efficient-Pytorch/releases/download/1.0/efficientdet-d7.pth", filename=self.system_dict["params"]["load_weights"])        


        # load last weights
        if self.system_dict["params"]["load_weights"] is not None:
            if self.system_dict["params"]["load_weights"].endswith('.pth'):
                weights_path = self.system_dict["params"]["load_weights"]
            else:
                weights_path = get_last_weights(self.system_dict["params"]["saved_path"])
            try:
                last_step = int(os.path.basename(weights_path).split('_')[-1].split('.')[0])
            except:
                last_step = 0


            try:
                ret = model.load_state_dict(torch.load(weights_path), strict=False)
            except RuntimeError as e:
                print(f'[Warning] Ignoring {e}')
                print(
                    '[Warning] Don\'t panic if you see this, this might be because you load a pretrained weights with different number of classes. The rest of the weights should be loaded already.')

            print(f'[Info] loaded weights: {os.path.basename(weights_path)}, resuming checkpoint from step: {last_step}')
        else:
            last_step = 0
            print('[Info] initializing weights...')
            init_weights(model)

        print("");
        print("");

        # freeze backbone if train head_only
        if self.system_dict["params"]["head_only"]:
            def freeze_backbone(m):
                classname = m.__class__.__name__
                for ntl in ['EfficientNet', 'BiFPN']:
                    if ntl in classname:
                        for param in m.parameters():
                            param.requires_grad = False

            model.apply(freeze_backbone)
            print('[Info] freezed backbone')

        print("");
        print("");

        if self.system_dict["params"]["num_gpus"] > 1 and self.system_dict["params"]["batch_size"] // self.system_dict["params"]["num_gpus"] < 4:
            model.apply(replace_w_sync_bn)
            use_sync_bn = True
        else:
            use_sync_bn = False

        writer = SummaryWriter(self.system_dict["params"]["log_path"] + f'/{datetime.datetime.now().strftime("%Y%m%d-%H%M%S")}/')

        model = ModelWithLoss(model, debug=self.system_dict["params"]["debug"])

        if self.system_dict["params"]["num_gpus"] > 0:
            model = model.cuda()
            if self.system_dict["params"]["num_gpus"] > 1:
                model = CustomDataParallel(model, self.system_dict["params"]["num_gpus"])
                if use_sync_bn:
                    patch_replication_callback(model)

        if self.system_dict["params"]["optim"] == 'adamw':
            optimizer = torch.optim.AdamW(model.parameters(), self.system_dict["params"]["lr"])
        else:
            optimizer = torch.optim.SGD(model.parameters(), self.system_dict["params"]["lr"], momentum=0.9, nesterov=True)

        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, patience=3, verbose=True)

        epoch = 0
        best_loss = 1e5
        best_epoch = 0
        step = max(0, last_step)
        model.train()

        num_iter_per_epoch = len(training_generator)


        try:
            for epoch in range(self.system_dict["params"]["num_epochs"]):
                last_epoch = step // num_iter_per_epoch
                if epoch < last_epoch:
                    continue

                epoch_loss = []
                progress_bar = tqdm(training_generator)
                for iter, data in enumerate(progress_bar):
                    if iter < step - last_epoch * num_iter_per_epoch:
                        progress_bar.update()
                        continue
                    try:
                        imgs = data['img']
                        annot = data['annot']

                        if self.system_dict["params"]["num_gpus"] == 1:
                            # if only one gpu, just send it to cuda:0
                            # elif multiple gpus, send it to multiple gpus in CustomDataParallel, not here
                            imgs = imgs.cuda()
                            annot = annot.cuda()

                        optimizer.zero_grad()
                        cls_loss, reg_loss = model(imgs, annot, obj_list=self.system_dict["params"]["obj_list"])
                        cls_loss = cls_loss.mean()
                        reg_loss = reg_loss.mean()

                        loss = cls_loss + reg_loss
                        if loss == 0 or not torch.isfinite(loss):
                            continue

                        loss.backward()
                        # torch.nn.utils.clip_grad_norm_(model.parameters(), 0.1)
                        optimizer.step()

                        epoch_loss.append(float(loss))

                        progress_bar.set_description(
                            'Step: {}. Epoch: {}/{}. Iteration: {}/{}. Cls loss: {:.5f}. Reg loss: {:.5f}. Total loss: {:.5f}'.format(
                                step, epoch, self.system_dict["params"]["num_epochs"], iter + 1, num_iter_per_epoch, cls_loss.item(),
                                reg_loss.item(), loss.item()))
                        writer.add_scalars('Loss', {'train': loss}, step)
                        writer.add_scalars('Regression_loss', {'train': reg_loss}, step)
                        writer.add_scalars('Classfication_loss', {'train': cls_loss}, step)

                        # log learning_rate
                        current_lr = optimizer.param_groups[0]['lr']
                        writer.add_scalar('learning_rate', current_lr, step)

                        step += 1

                        if step % self.system_dict["params"]["save_interval"] == 0 and step > 0:
                            self.save_checkpoint(model, f'efficientdet-d{self.system_dict["params"]["compound_coef"]}_trained.pth')
                            #print('checkpoint...')

                    except Exception as e:
                        print('[Error]', traceback.format_exc())
                        print(e)
                        continue
                scheduler.step(np.mean(epoch_loss))

                if epoch % self.system_dict["params"]["val_interval"] == 0 and self.system_dict["dataset"]["val"]["status"]:
                    print("Running validation");
                    model.eval()
                    loss_regression_ls = []
                    loss_classification_ls = []
                    for iter, data in enumerate(val_generator):
                        with torch.no_grad():
                            imgs = data['img']
                            annot = data['annot']

                            if self.system_dict["params"]["num_gpus"] == 1:
                                imgs = imgs.cuda()
                                annot = annot.cuda()

                            cls_loss, reg_loss = model(imgs, annot, obj_list=self.system_dict["params"]["obj_list"])
                            cls_loss = cls_loss.mean()
                            reg_loss = reg_loss.mean()

                            loss = cls_loss + reg_loss
                            if loss == 0 or not torch.isfinite(loss):
                                continue

                            loss_classification_ls.append(cls_loss.item())
                            loss_regression_ls.append(reg_loss.item())

                    cls_loss = np.mean(loss_classification_ls)
                    reg_loss = np.mean(loss_regression_ls)
                    loss = cls_loss + reg_loss

                    print(
                        'Val. Epoch: {}/{}. Classification loss: {:1.5f}. Regression loss: {:1.5f}. Total loss: {:1.5f}'.format(
                            epoch, self.system_dict["params"]["num_epochs"], cls_loss, reg_loss, loss))
                    writer.add_scalars('Loss', {'val': loss}, step)
                    writer.add_scalars('Regression_loss', {'val': reg_loss}, step)
                    writer.add_scalars('Classfication_loss', {'val': cls_loss}, step)

                    if loss + self.system_dict["params"]["es_min_delta"] < best_loss:
                        best_loss = loss
                        best_epoch = epoch

                        self.save_checkpoint(model, f'efficientdet-d{self.system_dict["params"]["compound_coef"]}_trained.pth')

                    model.train()

                    # Early stopping
                    if epoch - best_epoch > self.system_dict["params"]["es_patience"] > 0:
                        print('[Info] Stop training at epoch {}. The lowest loss achieved is {}'.format(epoch, best_loss))
                        break
        except KeyboardInterrupt:
            self.save_checkpoint(model, f'efficientdet-d{self.system_dict["params"]["compound_coef"]}_trained.pth')
            writer.close()
        writer.close()


        print("");
        print("");
        print("Training complete");
        modelPath = os.path.join(self.system_dict["params"]["saved_path"], f'efficientdet-d{self.system_dict["params"]["compound_coef"]}_trained.pth')
        print(f"Model saved to {modelPath}")





    def save_checkpoint(self, model, name):
        if isinstance(model, CustomDataParallel):
            torch.save(model.module.model.state_dict(), os.path.join(self.system_dict["params"]["saved_path"], name))
        else:
            torch.save(model.model.state_dict(), os.path.join(self.system_dict["params"]["saved_path"], name))


import torch
from torch.autograd import Variable
from torch import nn

class Normalize(nn.Module):

    def __init__(self, power=2):
        super(Normalize, self).__init__()
        self.power = power
    
    def forward(self, x):
        norm = x.pow(self.power).sum(1, keepdim=True).pow(1./self.power)
        out = x.div(norm)
        return out



import numpy as np
import scipy
import scipy.ndimage
from scipy.ndimage.filters import gaussian_filter
from scipy.ndimage.interpolation import map_coordinates
import collections
from PIL import Image
import numbers
import random

__author__ = "Wei OUYANG"
__license__ = "GPL"
__version__ = "0.1.0"
__status__ = "Development"


def center_crop(x, center_crop_size):
    assert x.ndim == 3
    centerw, centerh = x.shape[1] // 2, x.shape[2] // 2
    halfw, halfh = center_crop_size[0] // 2, center_crop_size[1] // 2
    return x[:, centerw - halfw:centerw + halfw, centerh - halfh:centerh + halfh]


def to_tensor(x):
    import torch
    x = x.transpose((2, 0, 1))
    return torch.from_numpy(x).float()


def random_num_generator(config, random_state=np.random):
    if config[0] == 'uniform':
        ret = random_state.uniform(config[1], config[2], 1)[0]
    elif config[0] == 'lognormal':
        ret = random_state.lognormal(config[1], config[2], 1)[0]
    else:
        print(config)
        raise Exception('unsupported format')
    return ret


def poisson_downsampling(image, peak, random_state=np.random):
    if not isinstance(image, np.ndarray):
        imgArr = np.array(image, dtype='float32')
    else:
        imgArr = image.astype('float32')
    Q = imgArr.max(axis=(0, 1)) / peak
    if Q[0] == 0:
        return imgArr
    ima_lambda = imgArr / Q
    noisy_img = random_state.poisson(lam=ima_lambda)
    return noisy_img.astype('float32')


def elastic_transform(image, alpha=1000, sigma=30, spline_order=1, mode='nearest', random_state=np.random):
    """Elastic deformation of image as described in [Simard2003]_.
    .. [Simard2003] Simard, Steinkraus and Platt, "Best Practices for
       Convolutional Neural Networks applied to Visual Document Analysis", in
       Proc. of the International Conference on Document Analysis and
       Recognition, 2003.
    """
    assert image.ndim == 3
    shape = image.shape[:2]

    dx = gaussian_filter((random_state.rand(*shape) * 2 - 1),
                         sigma, mode="constant", cval=0) * alpha
    dy = gaussian_filter((random_state.rand(*shape) * 2 - 1),
                         sigma, mode="constant", cval=0) * alpha

    x, y = np.meshgrid(np.arange(shape[0]), np.arange(shape[1]), indexing='ij')
    indices = [np.reshape(x + dx, (-1, 1)), np.reshape(y + dy, (-1, 1))]
    result = np.empty_like(image)
    for i in range(image.shape[2]):
        result[:, :, i] = map_coordinates(
            image[:, :, i], indices, order=spline_order, mode=mode).reshape(shape)
    return result


class Merge(object):
    """Merge a group of images
    """

    def __init__(self, axis=-1):
        self.axis = axis

    def __call__(self, images):
        if isinstance(images, collections.Sequence) or isinstance(images, np.ndarray):
            assert all([isinstance(i, np.ndarray)
                        for i in images]), 'only numpy array is supported'
            shapes = [list(i.shape) for i in images]
            for s in shapes:
                s[self.axis] = None
            assert all([s == shapes[0] for s in shapes]
                       ), 'shapes must be the same except the merge axis'
            return np.concatenate(images, axis=self.axis)
        else:
            raise Exception("obj is not a sequence (list, tuple, etc)")


class Split(object):
    """Split images into individual arraies
    """

    def __init__(self, *slices, **kwargs):
        assert isinstance(slices, collections.Sequence)
        slices_ = []
        for s in slices:
            if isinstance(s, collections.Sequence):
                slices_.append(slice(*s))
            else:
                slices_.append(s)
        assert all([isinstance(s, slice) for s in slices_]
                   ), 'slices must be consist of slice instances'
        self.slices = slices_
        self.axis = kwargs.get('axis', -1)

    def __call__(self, image):
        if isinstance(image, np.ndarray):
            ret = []
            for s in self.slices:
                sl = [slice(None)] * image.ndim
                sl[self.axis] = s
                ret.append(image[sl])
            return ret
        else:
            raise Exception("obj is not an numpy array")


class ElasticTransform(object):
    """Apply elastic transformation on a numpy.ndarray (H x W x C)
    """

    def __init__(self, alpha, sigma):
        self.alpha = alpha
        self.sigma = sigma

    def __call__(self, image):
        if isinstance(self.alpha, collections.Sequence):
            alpha = random_num_generator(self.alpha)
        else:
            alpha = self.alpha
        if isinstance(self.sigma, collections.Sequence):
            sigma = random_num_generator(self.sigma)
        else:
            sigma = self.sigma
        return elastic_transform(image, alpha=alpha, sigma=sigma)


class PoissonSubsampling(object):
    """Poisson subsampling on a numpy.ndarray (H x W x C)
    """

    def __init__(self, peak, random_state=np.random):
        self.peak = peak
        self.random_state = random_state

    def __call__(self, image):
        if isinstance(self.peak, collections.Sequence):
            peak = random_num_generator(
                self.peak, random_state=self.random_state)
        else:
            peak = self.peak
        return poisson_downsampling(image, peak, random_state=self.random_state)


class AddGaussianNoise(object):
    """Add gaussian noise to a numpy.ndarray (H x W x C)
    """

    def __init__(self, mean, sigma, random_state=np.random):
        self.sigma = sigma
        self.mean = mean
        self.random_state = random_state

    def __call__(self, image):
        if isinstance(self.sigma, collections.Sequence):
            sigma = random_num_generator(
                self.sigma, random_state=self.random_state)
        else:
            sigma = self.sigma
        if isinstance(self.mean, collections.Sequence, random_state=self.random_state):
            mean = random_num_generator(self.mean)
        else:
            mean = self.mean
        row, col, ch = image.shape
        gauss = self.random_state.normal(mean, sigma, (row, col, ch))
        gauss = gauss.reshape(row, col, ch)
        image += gauss
        return image


class AddSpeckleNoise(object):
    """Add speckle noise to a numpy.ndarray (H x W x C)
    """

    def __init__(self, mean, sigma, random_state=np.random):
        self.sigma = sigma
        self.mean = mean
        self.random_state = random_state

    def __call__(self, image):
        if isinstance(self.sigma, collections.Sequence):
            sigma = random_num_generator(
                self.sigma, random_state=self.random_state)
        else:
            sigma = self.sigma
        if isinstance(self.mean, collections.Sequence):
            mean = random_num_generator(
                self.mean, random_state=self.random_state)
        else:
            mean = self.mean
        row, col, ch = image.shape
        gauss = self.random_state.normal(mean, sigma, (row, col, ch))
        gauss = gauss.reshape(row, col, ch)
        image += image * gauss
        return image


class RandomGaussianBlurring(object):
    """Apply gaussian blur to a numpy.ndarray (H x W x C)
    """

    def __init__(self, sigma, p=0.2, random_state=np.random):
        self.sigma = sigma
        self.p = p
        self.random_state = random_state

    def __call__(self, image):
        if isinstance(self.sigma, collections.Sequence):
            sigma = random_num_generator(
                self.sigma, random_state=self.random_state)
        else:
            sigma = self.sigma
        if random.random() < self.p:
            image = gaussian_filter(image, sigma=(sigma, sigma, 0))
        return image


class AddGaussianPoissonNoise(object):
    """Add poisson noise with gaussian blurred image to a numpy.ndarray (H x W x C)
    """

    def __init__(self, sigma, peak, random_state=np.random):
        self.sigma = sigma
        self.peak = peak
        self.random_state = random_state

    def __call__(self, image):
        if isinstance(self.sigma, collections.Sequence):
            sigma = random_num_generator(
                self.sigma, random_state=self.random_state)
        else:
            sigma = self.sigma
        if isinstance(self.peak, collections.Sequence):
            peak = random_num_generator(
                self.peak, random_state=self.random_state)
        else:
            peak = self.peak
        bg = gaussian_filter(image, sigma=(sigma, sigma, 0))
        bg = poisson_downsampling(
            bg, peak=peak, random_state=self.random_state)
        return image + bg


class MaxScaleNumpy(object):
    """scale with max and min of each channel of the numpy array i.e.
    channel = (channel - mean) / std
    """

    def __init__(self, range_min=0.0, range_max=1.0):
        self.scale = (range_min, range_max)

    def __call__(self, image):
        mn = image.min(axis=(0, 1))
        mx = image.max(axis=(0, 1))
        return self.scale[0] + (image - mn) * (self.scale[1] - self.scale[0]) / (mx - mn)


class MedianScaleNumpy(object):
    """Scale with median and mean of each channel of the numpy array i.e.
    channel = (channel - mean) / std
    """

    def __init__(self, range_min=0.0, range_max=1.0):
        self.scale = (range_min, range_max)

    def __call__(self, image):
        mn = image.min(axis=(0, 1))
        md = np.median(image, axis=(0, 1))
        return self.scale[0] + (image - mn) * (self.scale[1] - self.scale[0]) / (md - mn)


class NormalizeNumpy(object):
    """Normalize each channel of the numpy array i.e.
    channel = (channel - mean) / std
    """

    def __call__(self, image):
        image -= image.mean(axis=(0, 1))
        s = image.std(axis=(0, 1))
        s[s == 0] = 1.0
        image /= s
        return image


class MutualExclude(object):
    """Remove elements from one channel
    """

    def __init__(self, exclude_channel, from_channel):
        self.from_channel = from_channel
        self.exclude_channel = exclude_channel

    def __call__(self, image):
        mask = image[:, :, self.exclude_channel] > 0
        image[:, :, self.from_channel][mask] = 0
        return image


class RandomCropNumpy(object):
    """Crops the given numpy array at a random location to have a region of
    the given size. size can be a tuple (target_height, target_width)
    or an integer, in which case the target will be of a square shape (size, size)
    """

    def __init__(self, size, random_state=np.random):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
        self.random_state = random_state

    def __call__(self, img):
        w, h = img.shape[:2]
        th, tw = self.size
        if w == tw and h == th:
            return img

        x1 = self.random_state.randint(0, w - tw)
        y1 = self.random_state.randint(0, h - th)
        return img[x1:x1 + tw, y1: y1 + th, :]


class CenterCropNumpy(object):
    """Crops the given numpy array at the center to have a region of
    the given size. size can be a tuple (target_height, target_width)
    or an integer, in which case the target will be of a square shape (size, size)
    """

    def __init__(self, size):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def __call__(self, img):
        w, h = img.shape[:2]
        th, tw = self.size
        x1 = int(round((w - tw) / 2.))
        y1 = int(round((h - th) / 2.))
        return img[x1:x1 + tw, y1: y1 + th, :]


class RandomRotate(object):
    """Rotate a PIL.Image or numpy.ndarray (H x W x C) randomly
    """

    def __init__(self, angle_range=(0.0, 360.0), axes=(0, 1), mode='reflect', random_state=np.random):
        assert isinstance(angle_range, tuple)
        self.angle_range = angle_range
        self.random_state = random_state
        self.axes = axes
        self.mode = mode

    def __call__(self, image):
        angle = self.random_state.uniform(
            self.angle_range[0], self.angle_range[1])
        if isinstance(image, np.ndarray):
            mi, ma = image.min(), image.max()
            image = scipy.ndimage.interpolation.rotate(
                image, angle, reshape=False, axes=self.axes, mode=self.mode)
            return np.clip(image, mi, ma)
        elif isinstance(image, Image.Image):
            return image.rotate(angle)
        else:
            raise Exception('unsupported type')


class BilinearResize(object):
    """Resize a PIL.Image or numpy.ndarray (H x W x C)
    """

    def __init__(self, zoom):
        self.zoom = [zoom, zoom, 1]

    def __call__(self, image):
        if isinstance(image, np.ndarray):
            return scipy.ndimage.interpolation.zoom(image, self.zoom)
        elif isinstance(image, Image.Image):
            return image.resize(self.size, Image.BILINEAR)
        else:
            raise Exception('unsupported type')


class EnhancedCompose(object):
    """Composes several transforms together.
    Args:
        transforms (List[Transform]): list of transforms to compose.
    Example:
        >>> transforms.Compose([
        >>>     transforms.CenterCrop(10),
        >>>     transforms.ToTensor(),
        >>> ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, img):
        for t in self.transforms:
            if isinstance(t, collections.Sequence):
                assert isinstance(img, collections.Sequence) and len(img) == len(
                    t), "size of image group and transform group does not fit"
                tmp_ = []
                for i, im_ in enumerate(img):
                    if callable(t[i]):
                        tmp_.append(t[i](im_))
                    else:
                        tmp_.append(im_)
                img = tmp_
            elif callable(t):
                img = t(img)
            elif t is None:
                continue
            else:
                raise Exception('unexpected type')
        return img




'''
Created on Nov 14, 2019

@author: Sascha Hornauer
'''
from MDistance import Distance_Eval
from PIL import Image
import os
import pandas as pd
import numpy as np
import cv2

def calc_to_csv(root_path,csv_filename,evaluator):
    
    distances = []
    
    for root, dirs, files in os.walk(root_path,followlinks=True):
        print ("Traversing {}".format(root))
        for file_ in files:
            filename = os.path.join(root, file_) 
            print (file_)
            if filename.endswith('jpg'):
                with open(filename, 'rb') as f:
                    img = Image.open(f)                    
                    distance, features = evaluator.get_distance_and_features(img)
                    distances.append({'filename':filename, 'distance':distance, 'features':features.squeeze()})            

    distances_pd = pd.DataFrame(distances)
    distances_pd.to_csv(csv_filename, sep=',',index=False)
    
if __name__ == '__main__':

    # Checkpoint
    evaluator = Distance_Eval('/home/saschaho/Simcenter/lemniscate.pytorch/checkpoint.pth.tar')
    #evaluator = Distance_Eval('/home/saschaho/Simcenter/lemniscate.pytorch/lemniscate_resnet50.pth.tar')

    csv_filename = '/home/saschaho/Simcenter/lemniscate.pytorch/distances.csv'
    
    root_path = '/home/saschaho/Simcenter/copy_for_zhirong/Florida/'
    recalculate_distances = True
    if recalculate_distances: calc_to_csv(root_path,csv_filename,evaluator)
    
    pkl_save_path = '/home/saschaho/Simcenter/copy_for_zhirong/pkl_stat_data'
    pkl_stats_file = '/home/saschaho/Simcenter/copy_for_zhirong/stats.pkl'
    
    ### Calculate and show distance images
    distances = pd.read_csv(csv_filename)
     
    closest_ten = distances.sort_values(by='distance').iloc[0:50]
    furthest_ten = distances.sort_values(by='distance').iloc[-70:]
     
    for i, image in enumerate(furthest_ten.iterrows()):
        image = image[1] # Some index number is at position 0
        img_path = image['filename']
        img_name = os.path.basename(img_path).split('.')[0]
         
        print ("Processing {}".format(img_path))
        img_default_dist = image['distance']
                          
        with open(img_path, 'rb') as f:
            img = Image.open(f)
            img_array = np.array(img)

        final_img = cv2.cvtColor(img_array, cv2.COLOR_RGB2BGR)
        #dist_at_pos_pd = evaluator.image_check(img_array, mode='peephole')
         
        #final_img = annotate_img(img_array=img_array, dist_at_pos_pd=dist_at_pos_pd, default_dist=img_default_dist)
        #cv2.imwrite('furthest_{}.png'.format(i+50),final_img)
        cv2.imshow('test',final_img)
        cv2.waitKey(1000)
        print (i)
          
        



from lib.train_detector import Detector
import argparse

def get_args():
    parser = argparse.ArgumentParser('EfficientDet-based number of floor detection model')
    parser.add_argument('-c', '--compound_coef', type=int, default=4,
                        help='Compund coefficient for the EfficientDet backbone, e.g., enter 7 for EfficientDet-D7') 
    parser.add_argument('-n', '--num_workers', type=int, default=0,
                        help='Number of workers of Dataloader')
    parser.add_argument('--head_only', type=boolean_string, default=False,
                        help='True if desired to finetune the regressor and the classifier only,'
                             'useful in early stage convergence or small/easy dataset')
    parser.add_argument('--num_gpus', type=int, default=1, 
                        help='Number of GPUs (Enter 0 for CPU-based training)')    
    parser.add_argument('--optim', type=str, default='adamw',
                        help='Select optimizer for training, '
                             'Use \'adamw\' until the last stage'
                             'then switch to \'sgd\'')
    parser.add_argument('--batch_size', type=int, default=2,
                        help='The number of images per batch')
    parser.add_argument('--lr', type=float, default=1e-4, help='Learning rate')
    parser.add_argument('--num_epochs', type=int, default=25, help='Number of training epochs')
    parser.add_argument('--data_path', type=str, default='datasets/',
                        help='Path for the root folder of dataset')
    parser.add_argument('--val_interval', type=int, default=1,
                        help='Number of epoches between validating phases')
    parser.add_argument('--save_interval', type=int, default=5,
                        help='Number of epoches between model saving')
    parser.add_argument('--es_min_delta', type=float, default=0.0,
                        help='Early stopping parameter: Minimum change in loss to qualify as an improvement')
    parser.add_argument('--es_patience', type=int, default=0,
                        help='Early stopping parameter: Number of epochs with no improvement after which training will be stopped. Set to 0 to disable this technique.')
    parser.add_argument('--customModel_path', type=str, 
                        default='models/efficientdet-d4_trained.pth',
                        help='Path for the custom pretrained model')

    args = parser.parse_args()
    return args

def boolean_string(s):
    if s not in {'False', 'True'}:
        raise ValueError('Not a valid boolean entry')
    return s == 'True'

def train(opt):
    # Create the Object Detector Object
    gtf = Detector()

    # Set the Training and Validation Datasets
    train_dir = "train"; val_dir = "valid"
    classes = ["floor"]
    
    gtf.set_train_dataset(opt.data_path,"","",train_dir,classes_list=classes,
                          batch_size=opt.batch_size, num_workers=opt.num_workers)
    
    gtf.set_val_dataset(opt.data_path,"","",val_dir)
    
    # Define the Model Architecture
    modelArchitecture = f"efficientdet-d{opt.compound_coef}.pth" ## Figure out the customModel_path
    
    gtf.set_model(model_name=modelArchitecture, num_gpus=opt.num_gpus,
                  freeze_head=opt.head_only)
    
    # Set Model Hyperparameters    
    gtf.set_hyperparams(optimizer=opt.optim, lr=opt.lr,
                        es_min_delta=opt.es_min_delta, es_patience=opt.es_patience)
    
    # Train    
    gtf.train(num_epochs=opt.num_epochs, val_interval=opt.val_interval,
              save_interval=opt.save_interval)

if __name__ == '__main__':
    opt = get_args()
    train(opt)


import os
import torch
import numpy as np

from torch.utils.data import Dataset, DataLoader
from pycocotools.coco import COCO
import cv2


class CocoDataset(Dataset):
    def __init__(self, root_dir, coco_dir, img_dir, set_dir='train2017', transform=None):

        self.root_dir = root_dir
        self.img_dir = img_dir
        self.coco_dir = coco_dir
        self.set_dir = set_dir
        self.set_name = set_dir
        self.transform = transform

        self.coco = COCO(os.path.join(self.root_dir, self.coco_dir, 'annotations', 'instances_' + self.set_name + '.json'))
        self.image_ids = self.coco.getImgIds()

        self.load_classes()
        
        #print(self.classes)
        #print(self.labels)

    def load_classes(self):

        # load class names (name -> label)
        categories = self.coco.loadCats(self.coco.getCatIds())
        categories.sort(key=lambda x: x['id'])

        self.classes = {}
        for c in categories:
            self.classes[c['name']] = len(self.classes)

        # also load the reverse (label -> name)
        self.labels = {}
        for key, value in self.classes.items():
            self.labels[value] = key

    def __len__(self):
        return len(self.image_ids)

    def __getitem__(self, idx):

        img = self.load_image(idx)
        annot = self.load_annotations(idx)
        sample = {'img': img, 'annot': annot}
        if self.transform:
            sample = self.transform(sample)
        return sample

    def load_image(self, image_index):
        image_info = self.coco.loadImgs(self.image_ids[image_index])[0]
        path = os.path.join(self.root_dir, self.coco_dir, self.set_dir, image_info['file_name'])
        img = cv2.imread(path)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        return img.astype(np.float32) / 255.

    def load_annotations(self, image_index):
        # get ground truth annotations
        annotations_ids = self.coco.getAnnIds(imgIds=self.image_ids[image_index], iscrowd=False)
        annotations = np.zeros((0, 5))

        # some images appear to miss annotations
        if len(annotations_ids) == 0:
            return annotations

        # parse annotations
        coco_annotations = self.coco.loadAnns(annotations_ids)
        for idx, a in enumerate(coco_annotations):

            # some annotations have basically no width / height, skip them
            if a['bbox'][2] < 1 or a['bbox'][3] < 1:
                continue

            annotation = np.zeros((1, 5))
            annotation[0, :4] = a['bbox']
            #annotation[0, 4] = a['category_id'] - 1
            annotation[0, 4] = a['category_id']
            annotations = np.append(annotations, annotation, axis=0)

        # transform from [x, y, w, h] to [x1, y1, x2, y2]
        annotations[:, 2] = annotations[:, 0] + annotations[:, 2]
        annotations[:, 3] = annotations[:, 1] + annotations[:, 3]

        return annotations


def collater(data):
    imgs = [s['img'] for s in data]
    annots = [s['annot'] for s in data]
    scales = [s['scale'] for s in data]

    imgs = torch.from_numpy(np.stack(imgs, axis=0))

    max_num_annots = max(annot.shape[0] for annot in annots)

    if max_num_annots > 0:

        annot_padded = torch.ones((len(annots), max_num_annots, 5)) * -1

        for idx, annot in enumerate(annots):
            if annot.shape[0] > 0:
                annot_padded[idx, :annot.shape[0], :] = annot
    else:
        annot_padded = torch.ones((len(annots), 1, 5)) * -1

    imgs = imgs.permute(0, 3, 1, 2)

    return {'img': imgs, 'annot': annot_padded, 'scale': scales}


class Resizer(object):
    """Convert ndarrays in sample to Tensors."""
    
    def __init__(self, img_size=512):
        self.img_size = img_size

    def __call__(self, sample):
        image, annots = sample['img'], sample['annot']
        height, width, _ = image.shape
        if height > width:
            scale = self.img_size / height
            resized_height = self.img_size
            resized_width = int(width * scale)
        else:
            scale = self.img_size / width
            resized_height = int(height * scale)
            resized_width = self.img_size

        image = cv2.resize(image, (resized_width, resized_height), interpolation=cv2.INTER_LINEAR)

        new_image = np.zeros((self.img_size, self.img_size, 3))
        new_image[0:resized_height, 0:resized_width] = image

        annots[:, :4] *= scale

        return {'img': torch.from_numpy(new_image).to(torch.float32), 'annot': torch.from_numpy(annots), 'scale': scale}


class Augmenter(object):
    """Convert ndarrays in sample to Tensors."""

    def __call__(self, sample, flip_x=0.5):
        if np.random.rand() < flip_x:
            image, annots = sample['img'], sample['annot']
            image = image[:, ::-1, :]

            rows, cols, channels = image.shape

            x1 = annots[:, 0].copy()
            x2 = annots[:, 2].copy()

            x_tmp = x1.copy()

            annots[:, 0] = cols - x2
            annots[:, 2] = cols - x_tmp

            sample = {'img': image, 'annot': annots}

        return sample


class Normalizer(object):

    def __init__(self, mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]):
        self.mean = np.array([[mean]])
        self.std = np.array([[std]])

    def __call__(self, sample):
        image, annots = sample['img'], sample['annot']

        return {'img': ((image.astype(np.float32) - self.mean) / self.std), 'annot': annots}



from .folder import ImageFolderInstance, MaskFolderInstance, CombinedMaskDataset
from .cifar import CIFAR10Instance, CIFAR100Instance
from .mnist import MNISTInstance

__all__ = ('ImageFolderInstance', 'MaskFolderInstance', 'CombinedMaskDataset', 'MNISTInstance', 'CIFAR10Instance', 'CIFAR100Instance')




"""
This HRNet implementation is modified from the following repository:
https://github.com/HRNet/HRNet-Semantic-Segmentation
"""

import logging
import torch
import torch.nn as nn
import torch.nn.functional as F
from .utils import load_url
from csail_segmentation_tool.csail_seg.lib.nn import SynchronizedBatchNorm2d

BatchNorm2d = SynchronizedBatchNorm2d
BN_MOMENTUM = 0.1
logger = logging.getLogger(__name__)


__all__ = ['hrnetv2']


model_urls = {
    'hrnetv2': 'http://sceneparsing.csail.mit.edu/model/pretrained_resnet/hrnetv2_w48-imagenet.pth',
}


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.conv3 = nn.Conv2d(planes, planes * self.expansion, kernel_size=1,
                               bias=False)
        self.bn3 = BatchNorm2d(planes * self.expansion,
                               momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class HighResolutionModule(nn.Module):
    def __init__(self, num_branches, blocks, num_blocks, num_inchannels,
                 num_channels, fuse_method, multi_scale_output=True):
        super(HighResolutionModule, self).__init__()
        self._check_branches(
            num_branches, blocks, num_blocks, num_inchannels, num_channels)

        self.num_inchannels = num_inchannels
        self.fuse_method = fuse_method
        self.num_branches = num_branches

        self.multi_scale_output = multi_scale_output

        self.branches = self._make_branches(
            num_branches, blocks, num_blocks, num_channels)
        self.fuse_layers = self._make_fuse_layers()
        self.relu = nn.ReLU(inplace=True)

    def _check_branches(self, num_branches, blocks, num_blocks,
                        num_inchannels, num_channels):
        if num_branches != len(num_blocks):
            error_msg = 'NUM_BRANCHES({}) <> NUM_BLOCKS({})'.format(
                num_branches, len(num_blocks))
            logger.error(error_msg)
            raise ValueError(error_msg)

        if num_branches != len(num_channels):
            error_msg = 'NUM_BRANCHES({}) <> NUM_CHANNELS({})'.format(
                num_branches, len(num_channels))
            logger.error(error_msg)
            raise ValueError(error_msg)

        if num_branches != len(num_inchannels):
            error_msg = 'NUM_BRANCHES({}) <> NUM_INCHANNELS({})'.format(
                num_branches, len(num_inchannels))
            logger.error(error_msg)
            raise ValueError(error_msg)

    def _make_one_branch(self, branch_index, block, num_blocks, num_channels,
                         stride=1):
        downsample = None
        if stride != 1 or \
           self.num_inchannels[branch_index] != num_channels[branch_index] * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.num_inchannels[branch_index],
                          num_channels[branch_index] * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                BatchNorm2d(num_channels[branch_index] * block.expansion,
                            momentum=BN_MOMENTUM),
            )

        layers = []
        layers.append(block(self.num_inchannels[branch_index],
                            num_channels[branch_index], stride, downsample))
        self.num_inchannels[branch_index] = \
            num_channels[branch_index] * block.expansion
        for i in range(1, num_blocks[branch_index]):
            layers.append(block(self.num_inchannels[branch_index],
                                num_channels[branch_index]))

        return nn.Sequential(*layers)

    def _make_branches(self, num_branches, block, num_blocks, num_channels):
        branches = []

        for i in range(num_branches):
            branches.append(
                self._make_one_branch(i, block, num_blocks, num_channels))

        return nn.ModuleList(branches)

    def _make_fuse_layers(self):
        if self.num_branches == 1:
            return None

        num_branches = self.num_branches
        num_inchannels = self.num_inchannels
        fuse_layers = []
        for i in range(num_branches if self.multi_scale_output else 1):
            fuse_layer = []
            for j in range(num_branches):
                if j > i:
                    fuse_layer.append(nn.Sequential(
                        nn.Conv2d(num_inchannels[j],
                                  num_inchannels[i],
                                  1,
                                  1,
                                  0,
                                  bias=False),
                        BatchNorm2d(num_inchannels[i], momentum=BN_MOMENTUM)))
                elif j == i:
                    fuse_layer.append(None)
                else:
                    conv3x3s = []
                    for k in range(i-j):
                        if k == i - j - 1:
                            num_outchannels_conv3x3 = num_inchannels[i]
                            conv3x3s.append(nn.Sequential(
                                nn.Conv2d(num_inchannels[j],
                                          num_outchannels_conv3x3,
                                          3, 2, 1, bias=False),
                                BatchNorm2d(num_outchannels_conv3x3,
                                            momentum=BN_MOMENTUM)))
                        else:
                            num_outchannels_conv3x3 = num_inchannels[j]
                            conv3x3s.append(nn.Sequential(
                                nn.Conv2d(num_inchannels[j],
                                          num_outchannels_conv3x3,
                                          3, 2, 1, bias=False),
                                BatchNorm2d(num_outchannels_conv3x3,
                                            momentum=BN_MOMENTUM),
                                nn.ReLU(inplace=True)))
                    fuse_layer.append(nn.Sequential(*conv3x3s))
            fuse_layers.append(nn.ModuleList(fuse_layer))

        return nn.ModuleList(fuse_layers)

    def get_num_inchannels(self):
        return self.num_inchannels

    def forward(self, x):
        if self.num_branches == 1:
            return [self.branches[0](x[0])]

        for i in range(self.num_branches):
            x[i] = self.branches[i](x[i])

        x_fuse = []
        for i in range(len(self.fuse_layers)):
            y = x[0] if i == 0 else self.fuse_layers[i][0](x[0])
            for j in range(1, self.num_branches):
                if i == j:
                    y = y + x[j]
                elif j > i:
                    width_output = x[i].shape[-1]
                    height_output = x[i].shape[-2]
                    y = y + F.interpolate(
                        self.fuse_layers[i][j](x[j]),
                        size=(height_output, width_output),
                        mode='bilinear',
                        align_corners=False)
                else:
                    y = y + self.fuse_layers[i][j](x[j])
            x_fuse.append(self.relu(y))

        return x_fuse


blocks_dict = {
    'BASIC': BasicBlock,
    'BOTTLENECK': Bottleneck
}


class HRNetV2(nn.Module):
    def __init__(self, n_class, **kwargs):
        super(HRNetV2, self).__init__()
        extra = {
            'STAGE2': {'NUM_MODULES': 1, 'NUM_BRANCHES': 2, 'BLOCK': 'BASIC', 'NUM_BLOCKS': (4, 4), 'NUM_CHANNELS': (48, 96), 'FUSE_METHOD': 'SUM'},
            'STAGE3': {'NUM_MODULES': 4, 'NUM_BRANCHES': 3, 'BLOCK': 'BASIC', 'NUM_BLOCKS': (4, 4, 4), 'NUM_CHANNELS': (48, 96, 192), 'FUSE_METHOD': 'SUM'},
            'STAGE4': {'NUM_MODULES': 3, 'NUM_BRANCHES': 4, 'BLOCK': 'BASIC', 'NUM_BLOCKS': (4, 4, 4, 4), 'NUM_CHANNELS': (48, 96, 192, 384), 'FUSE_METHOD': 'SUM'},
            'FINAL_CONV_KERNEL': 1
            }

        # stem net
        self.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=2, padding=1,
                               bias=False)
        self.bn1 = BatchNorm2d(64, momentum=BN_MOMENTUM)
        self.conv2 = nn.Conv2d(64, 64, kernel_size=3, stride=2, padding=1,
                               bias=False)
        self.bn2 = BatchNorm2d(64, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)

        self.layer1 = self._make_layer(Bottleneck, 64, 64, 4)

        self.stage2_cfg = extra['STAGE2']
        num_channels = self.stage2_cfg['NUM_CHANNELS']
        block = blocks_dict[self.stage2_cfg['BLOCK']]
        num_channels = [
            num_channels[i] * block.expansion for i in range(len(num_channels))]
        self.transition1 = self._make_transition_layer([256], num_channels)
        self.stage2, pre_stage_channels = self._make_stage(
            self.stage2_cfg, num_channels)

        self.stage3_cfg = extra['STAGE3']
        num_channels = self.stage3_cfg['NUM_CHANNELS']
        block = blocks_dict[self.stage3_cfg['BLOCK']]
        num_channels = [
            num_channels[i] * block.expansion for i in range(len(num_channels))]
        self.transition2 = self._make_transition_layer(
            pre_stage_channels, num_channels)
        self.stage3, pre_stage_channels = self._make_stage(
            self.stage3_cfg, num_channels)

        self.stage4_cfg = extra['STAGE4']
        num_channels = self.stage4_cfg['NUM_CHANNELS']
        block = blocks_dict[self.stage4_cfg['BLOCK']]
        num_channels = [
            num_channels[i] * block.expansion for i in range(len(num_channels))]
        self.transition3 = self._make_transition_layer(
            pre_stage_channels, num_channels)
        self.stage4, pre_stage_channels = self._make_stage(
            self.stage4_cfg, num_channels, multi_scale_output=True)

    def _make_transition_layer(
            self, num_channels_pre_layer, num_channels_cur_layer):
        num_branches_cur = len(num_channels_cur_layer)
        num_branches_pre = len(num_channels_pre_layer)

        transition_layers = []
        for i in range(num_branches_cur):
            if i < num_branches_pre:
                if num_channels_cur_layer[i] != num_channels_pre_layer[i]:
                    transition_layers.append(nn.Sequential(
                        nn.Conv2d(num_channels_pre_layer[i],
                                  num_channels_cur_layer[i],
                                  3,
                                  1,
                                  1,
                                  bias=False),
                        BatchNorm2d(
                            num_channels_cur_layer[i], momentum=BN_MOMENTUM),
                        nn.ReLU(inplace=True)))
                else:
                    transition_layers.append(None)
            else:
                conv3x3s = []
                for j in range(i+1-num_branches_pre):
                    inchannels = num_channels_pre_layer[-1]
                    outchannels = num_channels_cur_layer[i] \
                        if j == i-num_branches_pre else inchannels
                    conv3x3s.append(nn.Sequential(
                        nn.Conv2d(
                            inchannels, outchannels, 3, 2, 1, bias=False),
                        BatchNorm2d(outchannels, momentum=BN_MOMENTUM),
                        nn.ReLU(inplace=True)))
                transition_layers.append(nn.Sequential(*conv3x3s))

        return nn.ModuleList(transition_layers)

    def _make_layer(self, block, inplanes, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                BatchNorm2d(planes * block.expansion, momentum=BN_MOMENTUM),
            )

        layers = []
        layers.append(block(inplanes, planes, stride, downsample))
        inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(inplanes, planes))

        return nn.Sequential(*layers)

    def _make_stage(self, layer_config, num_inchannels,
                    multi_scale_output=True):
        num_modules = layer_config['NUM_MODULES']
        num_branches = layer_config['NUM_BRANCHES']
        num_blocks = layer_config['NUM_BLOCKS']
        num_channels = layer_config['NUM_CHANNELS']
        block = blocks_dict[layer_config['BLOCK']]
        fuse_method = layer_config['FUSE_METHOD']

        modules = []
        for i in range(num_modules):
            # multi_scale_output is only used last module
            if not multi_scale_output and i == num_modules - 1:
                reset_multi_scale_output = False
            else:
                reset_multi_scale_output = True
            modules.append(
                HighResolutionModule(
                    num_branches,
                    block,
                    num_blocks,
                    num_inchannels,
                    num_channels,
                    fuse_method,
                    reset_multi_scale_output)
            )
            num_inchannels = modules[-1].get_num_inchannels()

        return nn.Sequential(*modules), num_inchannels

    def forward(self, x, return_feature_maps=False):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu(x)
        x = self.layer1(x)

        x_list = []
        for i in range(self.stage2_cfg['NUM_BRANCHES']):
            if self.transition1[i] is not None:
                x_list.append(self.transition1[i](x))
            else:
                x_list.append(x)
        y_list = self.stage2(x_list)

        x_list = []
        for i in range(self.stage3_cfg['NUM_BRANCHES']):
            if self.transition2[i] is not None:
                x_list.append(self.transition2[i](y_list[-1]))
            else:
                x_list.append(y_list[i])
        y_list = self.stage3(x_list)

        x_list = []
        for i in range(self.stage4_cfg['NUM_BRANCHES']):
            if self.transition3[i] is not None:
                x_list.append(self.transition3[i](y_list[-1]))
            else:
                x_list.append(y_list[i])
        x = self.stage4(x_list)

        # Upsampling
        x0_h, x0_w = x[0].size(2), x[0].size(3)
        x1 = F.interpolate(
            x[1], size=(x0_h, x0_w), mode='bilinear', align_corners=False)
        x2 = F.interpolate(
            x[2], size=(x0_h, x0_w), mode='bilinear', align_corners=False)
        x3 = F.interpolate(
            x[3], size=(x0_h, x0_w), mode='bilinear', align_corners=False)

        x = torch.cat([x[0], x1, x2, x3], 1)

        # x = self.last_layer(x)
        return [x]


def hrnetv2(pretrained=False, **kwargs):
    model = HRNetV2(n_class=1000, **kwargs)
    if pretrained:
        model.load_state_dict(load_url(model_urls['hrnetv2']), strict=False)

    return model



import torch
from torch.autograd import Function
from torch import nn
import math

class LinearAverageOp(Function):
    @staticmethod
    def forward(self, x, y, memory, params):
        T = params[0].item()
        batchSize = x.size(0)

        # inner product
        out = torch.mm(x.data, memory.t())
        out.div_(T) # batchSize * N
        
        self.save_for_backward(x, memory, y, params)

        return out

    @staticmethod
    def backward(self, gradOutput):
        x, memory, y, params = self.saved_tensors
        batchSize = gradOutput.size(0)
        T = params[0].item()
        momentum = params[1].item()
        
        # add temperature
        gradOutput.data.div_(T)

        # gradient of linear
        gradInput = torch.mm(gradOutput.data, memory)
        gradInput.resize_as_(x)

        # update the non-parametric data
        weight_pos = memory.index_select(0, y.data.view(-1)).resize_as_(x)
        weight_pos.mul_(momentum)
        weight_pos.add_(torch.mul(x.data, 1-momentum))
        w_norm = weight_pos.pow(2).sum(1, keepdim=True).pow(0.5)
        updated_weight = weight_pos.div(w_norm)
        memory.index_copy_(0, y, updated_weight)
        
        return gradInput, None, None, None

class LinearAverage(nn.Module):

    def __init__(self, inputSize, outputSize, T=0.07, momentum=0.5):
        super(LinearAverage, self).__init__()
        stdv = 1 / math.sqrt(inputSize)
        self.nLem = outputSize

        self.register_buffer('params',torch.tensor([T, momentum]));
        stdv = 1. / math.sqrt(inputSize/3)
        self.register_buffer('memory', torch.rand(outputSize, inputSize).mul_(2*stdv).add_(-stdv))

    def forward(self, x, y):
        out = LinearAverageOp.apply(x, y, self.memory, self.params)
        return out




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : batchnorm_reimpl.py
# Author : acgtyrant
# Date   : 11/01/2018
#
# This file is part of Synchronized-BatchNorm-PyTorch.
# https://github.com/vacancy/Synchronized-BatchNorm-PyTorch
# Distributed under MIT License.

import torch
import torch.nn as nn
import torch.nn.init as init

__all__ = ['BatchNorm2dReimpl']


class BatchNorm2dReimpl(nn.Module):
    """
    A re-implementation of batch normalization, used for testing the numerical
    stability.

    Author: acgtyrant
    See also:
    https://github.com/vacancy/Synchronized-BatchNorm-PyTorch/issues/14
    """
    def __init__(self, num_features, eps=1e-5, momentum=0.1):
        super().__init__()

        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum
        self.weight = nn.Parameter(torch.empty(num_features))
        self.bias = nn.Parameter(torch.empty(num_features))
        self.register_buffer('running_mean', torch.zeros(num_features))
        self.register_buffer('running_var', torch.ones(num_features))
        self.reset_parameters()

    def reset_running_stats(self):
        self.running_mean.zero_()
        self.running_var.fill_(1)

    def reset_parameters(self):
        self.reset_running_stats()
        init.uniform_(self.weight)
        init.zeros_(self.bias)

    def forward(self, input_):
        batchsize, channels, height, width = input_.size()
        numel = batchsize * height * width
        input_ = input_.permute(1, 0, 2, 3).contiguous().view(channels, numel)
        sum_ = input_.sum(1)
        sum_of_square = input_.pow(2).sum(1)
        mean = sum_ / numel
        sumvar = sum_of_square - sum_ * mean

        self.running_mean = (
                (1 - self.momentum) * self.running_mean
                + self.momentum * mean.detach()
        )
        unbias_var = sumvar / (numel - 1)
        self.running_var = (
                (1 - self.momentum) * self.running_var
                + self.momentum * unbias_var.detach()
        )

        bias_var = sumvar / numel
        inv_std = 1 / (bias_var + self.eps).pow(0.5)
        output = (
                (input_ - mean.unsqueeze(1)) * inv_std.unsqueeze(1) *
                self.weight.unsqueeze(1) + self.bias.unsqueeze(1))

        return output.view(channels, batchsize, height, width).permute(1, 0, 2, 3).contiguous()




# -*- coding: utf-8 -*-
"""
/*------------------------------------------------------*
|                         BRAILS                        |
|                                                       |
| Author: Charles Wang,  UC Berkeley, c_w@berkeley.edu  |
|                                                       |
| Date:    1/19/2021                                    |
*------------------------------------------------------*/
"""

import os
from pathlib import Path
import wget
import requests 
import json
import shapely
from sys import exit
from brails.utils.geoDicts import *
from shapely.geometry import Point, Polygon, MultiPolygon
from zipfile import ZipFile
import geopandas as gpd
import osmnx as ox
import json


#workDir = 'tmp'

#baseurl = 'https://nominatim.openstreetmap.org/search?city={address}&county={county}&state={state}&polygon_geojson=1&format=json'
baseurl = 'https://nominatim.openstreetmap.org/search?city={city}&county={county}&state={state}&country={country}&polygon_geojson=1&format=json'

def getPlaceBoundary(city='',county='',state='',country='',save=True,fileName='',workDir='tmp'):
    """Function for downloading a city boundary.

    Args:
        city (str): Name of the city or county.
        state (str): State of the city.
        save (bool): Save file locally.
        fileName (str): Path of the file.

    Returns:
        boundary (shapely.geometry.multipolygon.MultiPolygon): boundary
        fileName (str): path of the boundary file

    """
    r = requests.get(baseurl.format(city=city,county=county,state=state,country=country))
    if r.status_code==200:
        r = r.json()
        if r:
            btype = r[0]['geojson']['type']

            if btype in ['MultiPolygon']:
                pts_l = r[0]['geojson']['coordinates'][0]
                boundary=[]
                for pts in pts_l:
                    boundary.append(Polygon(pts))
                boundary = MultiPolygon(boundary)
            elif btype in ['Polygon']:
                boundary = r[0]['geojson']['coordinates'][0]
                boundary = Polygon(boundary)
            else:
                pts = r[0]['geojson']['coordinates'][0]
                boundary = Polygon(pts)

            # save file
            if save:
                if fileName=='': 
                    bfilename = f'{city}_{state}_{country}_boundary.geojson'.replace(' ','_')
                    fileName=f'{workDir}/{bfilename}'
                with open(fileName, 'w') as outfile:
                    json.dump(shapely.geometry.mapping(boundary), outfile)
                print('{} saved.'.format(fileName))
            return boundary, fileName

        else:
            print(f'OSM didn\'t find {city} {county} {state} {country}. Exit.')
            return -1, -1
    else: 
        print('Network issue. Exit.')



def getMSFootprintsByPlace(place='',save=True,fileName='',workDir='tmp',GoogleMapAPIKey='',overwrite=False):
    """Function for downloading a building footprints.

    Args:
        place (str): Name of the city or county.
        save (bool): Save file locally.
        fileName (str): Path of the file.

    Returns:
        allFootprints (geopandas dataframe): Footprints
        fileName (str): Path to the footprint file.

    """

    if fileName=='': fileName = Path('{workDir}/{city}_footprints_MS.geojson'.format(workDir = workDir, city=place).replace(' ','_').replace(',','_'))

    if os.path.exists(fileName) and not overwrite:
        print('{} already exists.'.format(fileName))
        allFootprints = gpd.read_file(fileName)

    else:

        #stateName = stateNames[state.upper()]
        city, stateName, country = getStateNameByPlace(place=place,workDir=workDir,GoogleMapAPIKey=GoogleMapAPIKey)
        
        footprintFilename = Path('{}/{}'.format(workDir, footprintFiles[stateName].split('/')[-1]))

        # download
        if not footprintFilename.exists():
            print('Downloading footprints...')
            footprintFilename = wget.download(footprintFiles[stateName], out=workDir)
            print('{} downloaded'.format(footprintFilename))
        else: print('{} exists.'.format(footprintFilename))

        # unzip
        with ZipFile(footprintFilename, 'r') as zip: 
            footprintGeojsonFilename = Path('{}/{}'.format(workDir, zip.namelist()[0]))
            if not footprintGeojsonFilename.exists():

                print('Extracting all files now...') 
                zip.extractall(workDir) 
                print('Extraction done!') 
            else: print('{} exists.'.format(footprintGeojsonFilename))

        # slice
        placeBoundary, boundaryFilename = getPlaceBoundary(city=city,county='',state=stateName,country=country)#(place=place,state=stateName)
        if boundaryFilename==-1 or placeBoundary==-1:
            assert False, f"Couldn't find a boundary file for the place {place} from OSM with the parameters {city} {stateName} {country}"

        print(f'Loading all footprints from {boundaryFilename}...')
        gdf_mask = gpd.read_file(boundaryFilename)
        allFootprints = gpd.read_file(footprintGeojsonFilename,mask=gdf_mask,)

        # save
        if allFootprints.shape[0]<1:
            print('Didn\'t find footprints for this region.')
        else:
            allFootprints.to_file(fileName, driver='GeoJSON')
            print('Footprint saved at {}'.format(fileName))

    return allFootprints, fileName

def getStateNameByPlace(place='',workDir='tmp',GoogleMapAPIKey=''):
    url = f"https://maps.googleapis.com/maps/api/geocode/json?address={place}&key={GoogleMapAPIKey}"
    r = requests.get(url)
    place = place.replace(' ','_').replace(',','_')
    placeinfo = Path(f'{workDir}/{place}_placeinfo.json')
    f = open(placeinfo, 'wb')
    f.write(r.content)
    f.close()

    j = r.json()
    if j['status']=='OK':
        l = j['results'][0]['formatted_address'].split(',')
        print(l)
        country = l[-1].strip()
        stateCode = l[-2].strip()
        city = l[-3].strip()
        if country=='USA':
            return city, stateNames[stateCode], country
        else: return None,None,None

    else: 
        return None,None,None

def getStateNameByBBox(bbox,workDir='tmp',GoogleMapAPIKey=''):

    if len(bbox)!=4: 
        assert False, f"bbox should be a list with 4 elements. Your bbox {bbox} doesn't work."
        return None
    
    north = bbox[0]
    west = bbox[1]
    south = bbox[2]
    east = bbox[3]
    lon = (west+east)/2.
    lat = (north+south)/2.
    reverseURL = f'https://maps.googleapis.com/maps/api/geocode/json?latlng={lat},{lon}&key={GoogleMapAPIKey}'
    #print(reverseURL)

    r = requests.get(reverseURL)
    placeinfo = Path(f'{workDir}/{bbox[0]}_{bbox[1]}_{bbox[2]}_{bbox[3]}_placeinfo.json')
    f = open(placeinfo, 'wb')
    f.write(r.content)
    f.close()

    j = r.json()
    if j['status']=='OK':
        l = j['plus_code']['compound_code'].split(',')
        country = l[-1].strip()
        stateCode = l[-2].strip()
        if country=='USA':
            return stateNames[stateCode]
        else: return None

    else: 
        return None



def getMSFootprintsByBbox(bbox=[],stateName='',save=True,fileName='',workDir='tmp',overwrite=False,GoogleMapAPIKey=''):
    """Function for downloading a building footprints.

    Args:
        place (str): Name of the city or county.
        state (str): Abbreviation of state name
        save (bool): Save file locally.
        fileName (str): Path of the file.

    Returns:
        allFootprints (geopandas dataframe): Footprints
        fileName (str): Path to the footprint file.

    """

    if fileName=='': fileName = Path(f'{workDir}/{bbox[0]}_{bbox[1]}_{bbox[2]}_{bbox[3]}_footprints.geojson')

    if os.path.exists(fileName) and not overwrite:
        print('{} already exists.'.format(fileName))
        allFootprints = gpd.read_file(fileName)

    else:

        if stateName == '':
            #stateName = stateNames[state.upper()]
            stateName = getStateNameByBBox(bbox,workDir=workDir,GoogleMapAPIKey=GoogleMapAPIKey)
            if stateName is None:
                print('Didn\'t find the state name by the bbox you provided.')
                return None, None
        
        footprintFilename = Path('{}/{}'.format(workDir, footprintFiles[stateName].split('/')[-1]))

        # download
        if not footprintFilename.exists():
            print('Downloading footprints...')
            footprintFilename = wget.download(footprintFiles[stateName], out=workDir)
            print('{} downloaded'.format(footprintFilename))
        else: print('{} exists.'.format(footprintFilename))

        # unzip
        with ZipFile(footprintFilename, 'r') as zip: 
            footprintGeojsonFilename = Path('{}/{}'.format(workDir, zip.namelist()[0]))
            if not footprintGeojsonFilename.exists():

                print('Extracting all files now...') 
                zip.extractall(workDir) 
                print('Extraction done!') 
            else: print('{} exists.'.format(footprintGeojsonFilename))

        # slice
        print('Reading footprints file ...')
        allFootprints = gpd.read_file(footprintGeojsonFilename,bbox=(bbox[1],bbox[0],bbox[3],bbox[2]))

        # save
        if allFootprints.shape[0]<1:
            print('Didn\'t find footprints for this region.')
        else:
            allFootprints.to_file(fileName, driver='GeoJSON')
            print('Footprint saved at {}'.format(fileName))

    return allFootprints, fileName

def getOSMFootprints(bbox=[],place='',save=True,fileName='',workDir='tmp',overwrite=False):
    """Function for downloading a building footprints.

    Args:
        place (str): Name of the city or county.
        save (bool): Save file locally.
        fileName (str): Path of the file.

    Returns:
        allFootprints (geopandas dataframe): Footprints
        fileName (str): Path to the footprint file.

    """
    
    if fileName=='': 

        if len(bbox)>0: 
            fileName = Path(f'{workDir}/{bbox[0]}_{bbox[1]}_{bbox[2]}_{bbox[3]}_footprints.geojson')
        else:
            fileName = Path(f'{workDir}/{place}_footprints_OSM.geojson'.replace(' ','_').replace(',','_'))


    if os.path.exists(fileName) and not overwrite:
        print('{} already exists.'.format(fileName))
        allFootprints = gpd.read_file(fileName)

    else:
        if len(bbox)>0:
            allFootprints = ox.geometries_from_bbox(bbox[0],bbox[2],bbox[3],bbox[1],tags = {'building': True})# (north,west,south,east)->(north, south, east, west, tags)
        else:
            allFootprints = ox.geometries_from_place(place, tags = {'building': True})
        
        allFootprints = allFootprints[['geometry']]
        allFootprints = allFootprints[allFootprints['geometry'].type == 'Polygon']

        # save
        if allFootprints.shape[0]<1:
            print('Didn\'t find footprints for this region.')
        else:
            allFootprints.to_file(fileName, driver='GeoJSON')
            print('Footprint saved at {}'.format(fileName))

    return allFootprints, fileName



# -*- coding: utf-8 -*-
"""
/*------------------------------------------------------*
|                         BRAILS                        |
|                                                       |
| Author: Charles Wang,  UC Berkeley, c_w@berkeley.edu  |
|                                                       |
| Date:    1/10/2021                                    |
*------------------------------------------------------*/
"""

import os
import random
from multiprocessing.dummy import Pool as ThreadPool
import requests 
from pathlib import Path


def capturePic(browser, picname):
    try:
        localurl = browser.save_screenshot(picname)
        print("%s : Success" % localurl)
    except BaseException as msg:
        print("Fail：%s" % msg)


def download(urls):
    xcount = 0
    nlimit = 1e10
    reDownload = urls[0][6]

    for ls in urls:
        urlTop = ls[0]
        urlStreet = ls[1]
        lon = ls[2]
        lat = ls[3]
        cats = ls[4]
        imgDir = ls[5]
        #reDownload = ls[6]

        '''
        if rooftype not in roofcat:
            print('not in roofcat')
            continue
        '''

        #if not os.path.exists(thisFileDir):
        #    os.makedirs(thisFileDir)

        #numoffiles = len(os.listdir(thisFileDir))
        if xcount < nlimit: #numoffiles < maxNumofRoofImgs:

            for cat in cats:
                if cat == 'StreetView': trueURL = urlStreet
                elif cat == 'TopView': trueURL = urlTop
                #picname = thisFileDir + '/{prefix}x{lon}x{lat}.png'.format(prefix='StreetView',lon=lon,lat=lat)
                picname = Path(f'{imgDir}/{cat}/{cat}x{lon}x{lat}.png')

                exist = os.path.exists(picname)
                if not exist or (exist and reDownload):

                    r = requests.get(trueURL)
                    f = open(picname, 'wb')
                    f.write(r.content)
                    f.close()
                    xcount += 1

                    if os.path.getsize(picname)/1024 < 9: 
                        #print(urlStreet)
                        print(f"empty image from API: {lon}, {lat}")
                        #exit() # empty image from API

        else:
            break

# construct urls
def getGoogleImages(footprints=None, GoogleMapAPIKey='',imageTypes=['StreetView','TopView'],imgDir='',ncpu=2,fov=60,pitch=0,reDownloadImgs=False):

    if footprints is None:
        print('Please provide footprints') 
        return None

    if GoogleMapAPIKey == '':
        print('Please provide GoogleMapAPIKey') 
        return None

    for imgType in imageTypes:
        tmpImgeDir = os.path.join(imgDir, imgType)
        if not os.path.exists(tmpImgeDir): os.makedirs(tmpImgeDir)

    
    # APIs
    baseurl_streetview = "https://maps.googleapis.com/maps/api/streetview?size=640x640&location={lat},{lon}&fov={fov}&pitch={pitch}&source=outdoor&key="+GoogleMapAPIKey
    # consider using 256x256 to save disk
    
    baseurl_satellite="https://maps.googleapis.com/maps/api/staticmap?center={lat},{lon}&zoom=20&scale=1&size=256x256&maptype=satellite&key="+GoogleMapAPIKey+"&format=png&visual_refresh=true"

    #footprints = gpd.read_file(BuildingFootPrintsFileName)
    urls = []

    for ind, row in footprints.iterrows():
        o = row['geometry'].centroid
        lon, lat = '%.6f'%o.x, '%.6f'%o.y

        # a top view
        urlTop = baseurl_satellite.format(lat=lat,lon=lon)
        urlStreet = baseurl_streetview.format(lat=lat,lon=lon,fov=fov,pitch=pitch)
        cats = imageTypes
        reDownload = 1 if reDownloadImgs else 0
        urls.append([urlTop,urlStreet,lon,lat,cats,imgDir,reDownload])

    #print('shuffling...')
    #random.shuffle(urls)
    #print('shuffled...')     

    # divide urls into small chunks
    #ncpu = 4
    step = int(len(urls)/ncpu)+1
    chunks = [urls[x:x+step] for x in range(0, len(urls), step)]

    print('Downloading images from Google API ...')
    # get some workers
    pool = ThreadPool(ncpu)
    # send job to workers
    results = pool.map(download, chunks)
    # jobs are done, clean the site
    pool.close()
    pool.join()
    print('Images downloaded ...')


